/*

  Elmo driver for P0 and later hardware

  Copyright TiVo Inc., 2003-2004.   The terms of the GNU Public License 
  (version 2) apply to this code.

  Derived in part from g_NCR5380 and other Linux SCSI and IDE drivers.
  My thanks to the numerous authors of that code, who should in no way
  be held at all responsible for the various horrors I've visited upon
  the code thus abstracted.


*/

#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/mman.h>
#include <linux/pci.h>
#include <linux/linkage.h>
#include <linux/time.h>
#include <linux/wait.h>
#include <linux/blkdev.h>
#include <linux/blk.h>
#include <linux/spinlock.h>
#include <linux/tivoconfig.h>
#include <linux/list.h>
#include <linux/hdreg.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>

#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/semaphore.h>
#include <asm/bcm7315.h>
#include <asm/px_hw.h>

#include <scsi/elmointerface.h>

#include "scsi.h"
#include "hosts.h"
#include "constants.h"
#include "sd.h"

#include "elmo.h"

#include "ElmoPlog.h"

#if (NDEBUG & NDEBUG_LISTS)
#define LIST(x,y) {printk("LINE:%d   Adding %p to %p\n", __LINE__, (void*)(x), (void*)(y)); if ((x)==(y)) udelay(5); }
#define REMOVE(w,x,y,z) {printk("LINE:%d   Removing: %p->%p  %p->%p \n", __LINE__, (void*)(w), (void*)(x), (void*)(y), (void*)(z)); if ((x)==(y)) udelay(5); }
#else
#define LIST(x,y)
#define REMOVE(w,x,y,z)
#endif

#define KERN_DPLATT KERN_DEBUG

#define TAG_NEXT	-1
#define TAG_NONE	-2

#define USLEEP_SLEEP (20*HZ/1000)
#define USLEEP_WAITLONG USLEEP_SLEEP

#undef SIMULATE_IRQ

const static char version[] = BUILDDATE;

enum FakeCheckCondition {
  NothingPending=0,
  UnitAttention=1,
  IllegalRequest=2,
  NotReady=3,
  IllegalAddress=4,
};

/*
  Definitions for dealing with the 7315/7317
*/

#define PHYSADDR(base,offset) ((base) + (offset))

/*
  bit 4 set to 0 selects GPIO 33
  bit 5 set to 0 selects GPIO 3
*/
#define BCM_GPIO_SEL0        (PHYSADDR(BCM7315_REG_BASE,0x8007))
#define BCM_GPIO_SEL1        (PHYSADDR(BCM7315_REG_BASE,0x8006))
#define BCM_SPI_EBI_REQ_MUX  (PHYSADDR(BCM7315_REG_BASE,0x80A0))

/*
  PLOAD:
  GPIO_BASE + 0x3 bit 3 is GPIO3 data bit
  GPIO_BASE + 0x7 bit 3 is GPIO3 direction bit
  GPIO_BASE + 0x8 bit 3 is GPIO3 open drain bit

  PROGRAM:
  GPIO_BASE + 0x51 bit 1 is GPIO33 data bit
  GPIO_BASE + 0x53 bit 1 is GPIO33 direction bit
  GPIO_BASE + 0x55 bit 1 is GPIO33 open drain bit
*/

#define BCM_GPIO_BASE     (PHYSADDR(BCM7315_REG_BASE,0x50))

#define CHIP_ID      *((volatile unsigned long *)(PHYSADDR(BCM7315_REG_BASE, 0x0600)))

#define BCM_MTN_BASE            0xffe00000
#define MTN_RB_MMC_CLIENT_INDEX                  0x0000000c /* Memory Controller Client Index Address Register */
#define MTN_RB_MMC_INDX_DATA_WR                  0x00000010 /* Memory Controller Client Index Data Write Register */

#define BCM_MTN_CLIENT_INDEX    (volatile unsigned long *)(PHYSADDR(BCM_MTN_BASE, MTN_RB_MMC_CLIENT_INDEX))
#define BCM_MTN_INDEX_DATA_WR   (volatile unsigned long *)(PHYSADDR(BCM_MTN_BASE, MTN_RB_MMC_INDX_DATA_WR))
#define BCM_ISB_MMC_CLIENT_ID   0x12


#define TEI_BASE              ELMO_BRIDGE_BASE
#define TEI_CODE_LOAD         (PHYSADDR(TEI_BASE,0))
#define TEI_REVISION          (PHYSADDR(TEI_BASE,0x1C))

#define XMIT_ELMO_SPI_SUCCESS 0
#define XMIT_ELMO_SPI_FAILURE -1

#define MAX_ISB_BURST   64
#define ESS_ATA_TRANSFER_ABORT  0xF0

int xmit_elmo_spi (unsigned char *data, int len, void (*callback)(int));

static int elmo_abort(Scsi_Cmnd *);
static int elmo_detect(Scsi_Host_Template *);
static int elmo_release_resources(struct Scsi_Host *);
static int elmo_queue_command(Scsi_Cmnd *, void (*done) (Scsi_Cmnd *));
static int elmo_reset(Scsi_Cmnd *, unsigned int);
static const char *elmo_info(struct Scsi_Host *);
static int elmo_proc_info(char *buffer, char **start, off_t offset, int length, int hostno, int inout);
static int elmo_ioctl(Scsi_Device *dev, int cmd, void *arg);

static void elmo_intr(int irq, void *dev_id, struct pt_regs *regs);
static void do_elmo_intr(int irq, void *dev_id, struct pt_regs *regs);
static void broadcom_intr(int irq, int *useFlag, unsigned long softInterrupt);
static void broadcom_isb_intr(int irq, void *dev_id, struct pt_regs *regs);

static void elmo_wait (int mscount);

static int elmoCheckHostDMARules (Scsi_Cmnd * cmd);
static int elmoCheckMoreHostDMA (Scsi_Cmnd * cmd);
static int elmoTransferToUser (Scsi_Cmnd * cmd, unsigned char *buffer, unsigned int length);
static int elmoTransferFromUser (Scsi_Cmnd * cmd, unsigned char *buffer, unsigned int length);
static int elmoInquiry (Scsi_Cmnd * cmd, int device, int deviceType, char *vendor, char *model, char *version);
static void elmoCheckCondition (struct Scsi_Host *instance,
			      Scsi_Cmnd * cmd, enum FakeCheckCondition cond);
static void elmoRequestSense(struct Scsi_Host *instance,
			   Scsi_Cmnd * cmd);

static void run_main(void);
static void elmo_main(void);
static int elmo_probe_bridge(void);
static void elmo_bridge_is_alive(struct Scsi_Host *instance);
static void elmo_init_broadcom(void);
static void initialize_SCp(Scsi_Cmnd * cmd);
static int elmo_run_workers(struct Scsi_Host *instance);

static unsigned long elmo_interrupt_for_workers(struct Scsi_Host *instance,
					      unsigned long interruptsFired);

static void elmo_queue_worker_for_interrupt (struct elmo_worker *worker,
					   unsigned long interruptsNeeded,
					   enum elmo_worker_state nextState);

static void elmo_queue_worker_for_timer (struct elmo_worker *worker,
				       int timerTicks,
				       enum elmo_worker_state nextState);

static void elmo_worker_next_state (struct elmo_worker *worker,
			      enum elmo_worker_state nextState);

static void elmo_worker_disarm_timer (struct elmo_worker *worker);

static void elmo_worker_set_deadman (struct elmo_worker *worker,
				   int timerTicks);

static void elmo_worker_disarm_deadman (struct elmo_worker *worker);

static struct Scsi_Host *first_instance = NULL;
static Scsi_Host_Template *the_template = NULL;

static int elmoBridgeDispatchFunction (struct Scsi_Host *instance,
				     Scsi_Cmnd * cmd, int device);
static int elmoBridgeIoctlFunction (struct Scsi_Host *instance,
				  Scsi_Device *dev, int cmd, void *arg);

static void elmo_worker_setup(struct Scsi_Host *instance,
			      struct elmo_worker *worker);
static void elmo_worker_get_bulk_interface(struct Scsi_Host *instance,
					   struct elmo_worker *worker);
static void elmo_worker_get_ess_command(struct Scsi_Host *instance,
					struct elmo_worker *worker);
static void elmo_worker_setup_dma(struct Scsi_Host *instance,
				  struct elmo_worker *worker);
static void elmo_worker_await_fifo(struct Scsi_Host *instance,
				   struct elmo_worker *worker);
static void elmo_worker_finish_dma(struct Scsi_Host *instance,
				   struct elmo_worker *worker);
static void elmo_worker_done(struct Scsi_Host *instance,
			     struct elmo_worker *worker);

static void elmo_worker_service_read_fifo(struct Scsi_Host *instance,
					  struct elmo_worker *worker);
static void elmo_worker_service_write_fifo(struct Scsi_Host *instance,
					   struct elmo_worker *worker);
static void elmo_worker_gate_writer(struct Scsi_Host *instance,
				    struct elmo_worker *worker);
static void elmo_worker_become_writer(struct Scsi_Host *instance,
				      struct elmo_worker *worker);
static void elmo_worker_writer(struct Scsi_Host *instance,
			       struct elmo_worker *worker);
static void elmo_worker_become_reader(struct Scsi_Host *instance,
				      struct elmo_worker *worker);
static void elmo_worker_reader(struct Scsi_Host *instance,
			       struct elmo_worker *worker);

static void elmo_release_chip_resources (struct Scsi_Host *instance,
				       unsigned int resourceMask);

static void elmo_bug(struct elmo_worker *worker, 
		     const char *reason);

static void elmo_task (void *whatever);
static void elmo_timer_task (void *whatever);

static void show_elmo_intr(unsigned long);

int elmo_ll_open (struct inode *, struct file *);
ssize_t elmo_ll_read (struct file *, char *, size_t, loff_t *);
ssize_t elmo_ll_readv (struct file *filp, const struct iovec *iov, 
		       unsigned long count, loff_t *ppos);
ssize_t elmo_ll_write_common (struct file *, const char *, size_t, loff_t *,
			      int defer);
ssize_t elmo_ll_write (struct file *, const char *, size_t, loff_t *);
ssize_t elmo_ll_write_deferred (struct file *, const char *, size_t, loff_t *);
ssize_t elmo_ll_writev_common (struct file *filp, const struct iovec *iov, 
			       unsigned long count, loff_t *ppos, int defer);
ssize_t elmo_ll_writev (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *ppos);
ssize_t elmo_ll_writev_deferred (struct file *filp, const struct iovec *iov, 
				 unsigned long count, loff_t *ppos);
int elmo_ll_ioctl (struct inode *i, struct file *f, unsigned int cd,
		   unsigned long param);

#define ELMO_DEVICES 1

#define ELMO_DEVICE_BRIDGE 0

#define ELMO_HANDLERS 1

#define ELMO_HANDLER_BRIDGE 0

#define ELMO_MAX_PENDING_WRITERS 5

#define ELMO_BULK_FIFO_SIZE 512

typedef int (*ELMOCommandDispatchFunction) (struct Scsi_Host *instance,
					  Scsi_Cmnd * cmd, int device);

typedef int (*ELMOIoctlFunction) (struct Scsi_Host *instance,
				Scsi_Device *dev, int cmd, void *arg);

static const ELMOCommandDispatchFunction dispatchers[ELMO_HANDLERS] = 
  {
    elmoBridgeDispatchFunction,
  };

static const ELMOIoctlFunction ioctlHandlers[ELMO_HANDLERS] = 
  {
    elmoBridgeIoctlFunction,
  };

static const int deviceDispatcher[ELMO_DEVICES] = {0};

#define ELMO {							\
	proc_info:      elmo_proc_info,				\
	name:           "Elmo Scsi Driver",			\
	detect:         elmo_detect,				\
	release:        elmo_release_resources,			\
	info:           (void *)elmo_info,			\
        ioctl:          elmo_ioctl,				\
	queuecommand:   elmo_queue_command,			\
	abort:          elmo_abort,				\
	reset:          elmo_reset, 				\
	bios_param:     NULL,					\
	can_queue:      8,					\
	cmd_per_lun:    1,					\
        this_id:        7,					\
        sg_tablesize:   SG_ALL,					\
        use_clustering: DISABLE_CLUSTERING}

struct BroadcomDMADescriptor {
  int reserved                :  4;
  int transferByteCount       : 12;
  int ob                      :  1;
  int pd                      :  3;
  int l                       :  1;
  int f                       :  1;
  int w                       :  1;
  int pd1                     :  9;
  int memoryAddr              : 32;
};

static struct elmo_hostdata {
  unsigned char id_mask, id_higher_mask;	/* 1 << id, all bits greater */
  unsigned char targets_present;		/* targets we have connected
						   to, so we can call a select
						   failure a retryable condition */
  volatile unsigned char busy[ELMO_HANDLERS];	/* index = target, bit = lun */
  volatile unsigned char pending[ELMO_HANDLERS];	/* index = target, bit = lun */
  volatile unsigned char last_message;	/* last message OUT */
  volatile Scsi_Cmnd *connected[ELMO_HANDLERS];	/* currently connected command */
  volatile enum FakeCheckCondition condition[8]; /* pending conditions */
  volatile Scsi_Cmnd *issue_queue;	/* waiting to be issued */
  volatile unsigned aborted:1;		/* flag, says aborted */
  volatile unsigned pldLoaded:1;        /* flag, bridge is up */
  volatile unsigned initDone:1;         /* flag, initialization done */
  volatile unsigned atapiOK:1;          /* flag, ATAPI usable */
  volatile unsigned wasReset:1;         /* flag, ATAPI bus was reset, need to settle... */
  volatile unsigned mustReset:1;        /* flag, ATAPI reset is required */
  int flags;
  rwlock_t queueLock;                   /* lock on the issue queue */
  unsigned long time_expires;		/* in jiffies, set prior to sleeping */
  struct Scsi_Host *next_timer;
  int select_time;			/* timer in select for target response */
  volatile Scsi_Cmnd *selecting;
  int workerSN;
  struct list_head workers;
  volatile unsigned long resourcesBusy; /* On-chip resources */
  unsigned int atapiSectorSize;
  unsigned int atapiBufferModulus;
  int diagMode;
  struct BroadcomDMADescriptor *dmaDescriptors;
  unsigned char deferred_sense_buffer[SCSI_SENSE_BUFFERSIZE];
  int haveDeferredSense;
  unsigned char *llReadBuffer;
  unsigned int llReadBufferHead, llReadBufferTail;
  unsigned char *llWriteBuffer;
  unsigned int llWriteBufferHead, llWriteBufferTail;
  struct semaphore writeSem;
  unsigned int bulkFifoSize;
  int haveCommand;
  unsigned short ataCommandReg;
  unsigned short ataCylinderReg;
#ifdef ELMO_STATS
  unsigned timebase;			/* Base for time calcs */
  long time_read[8];			/* time to do reads */
  long time_write[8];			/* time to do writes */
  unsigned long bytes_read[8];		/* bytes read */
  unsigned long bytes_write[8];		/* bytes written */
  unsigned pendingr;
  unsigned pendingw;
#endif
} *thehostdata;

static struct file_operations elmo_ll_fops = {
  read:    elmo_ll_read,
  write:   elmo_ll_write,
  ioctl:   elmo_ll_ioctl,
  open:    elmo_ll_open,
  readv:   elmo_ll_readv,
  writev:  elmo_ll_writev,
};

static struct file_operations elmo_ll_deferred_fops = {
  read:   elmo_ll_read,
  write:  elmo_ll_write_deferred,
  ioctl:  elmo_ll_ioctl,
  open:   elmo_ll_open,
  readv:  elmo_ll_readv,
  writev: elmo_ll_writev_deferred,
};

#undef TEMPORARY_SLOW_DMA_HACK

static void elmo_reset_ATA(struct elmo_hostdata *hostdata);

int elmo_mem_lists_initialized = 0;
int elmo_probed_memsize = 0;

#ifdef SIMULATE_IRQ
static int runIRQs = 0;
#endif

static int elmo_broadcom_isb_enabled = 0;
static int elmo_asic_read_dma_enabled = 0;
static int elmo_asic_write_dma_enabled = 0;

static struct tq_struct elmo_task_struct;
static struct tq_struct elmo_timer_task_struct;
static wait_queue_head_t readerWaitQueue;

#if defined(MODULE)
MODULE_AUTHOR("Dave Platt, TiVo Inc.");
MODULE_DESCRIPTION("Elmo device driver");
MODULE_LICENSE("GPL");

MODULE_PARM(debug, "1-1i");
MODULE_PARM_DESC(debug,"enable debug loelmo");
static int debug = 0;

MODULE_PARM(jam, "1-10i");
MODULE_PARM_DESC(jam,"jam in a DMA failure occasionally");
static int jam = 0;

MODULE_PARM(autodebug, "0-600i");
MODULE_PARM_DESC(autodebug,"enable debug after DMA failure for N seconds");
static int autodebug = 0;

MODULE_PARM(autodebugcount, "0-999i");
MODULE_PARM_DESC(autodebugcount,"enable debug after DMA failure N times");
static int autodebugcount = 0;

MODULE_PARM(swaponread, "0-1i");
MODULE_PARM_DESC(swaponread,"enable byte swapping in FPGA for bulk data read");
static int swaponread=1;

MODULE_PARM(swaponwrite, "0-1i");
MODULE_PARM_DESC(swaponwrite,"enable byte swapping in FPGA for bulk data write");
static int swaponwrite=1;

static int autodebugExpires = 0;
static int autodebugRestore = -1;

#endif

static const char *bug_line1[4] =
{ "Bit-decay spreads through the Internet",
  "Data inconsistencies disrupt all nearby computers",
  "The home network crashes",
  "Video recordings are lost"};

static const char *bug_line2[4] =
{ "The global economy collapses",
  "PVRs stop working all across America",
  "No one knows who the Survivor is",
  "Riots ensue"};

static int notifier_installed = 0;
#define IS_7317A0()     (((CHIP_ID & 0xFFFFU) == 0) ? 1 : 0)

static void elmo_autodebug_on(void) {
  if (autodebug == 0 || debug > 1 || autodebugcount <= 0) {
    return;
  }
  if (autodebugRestore < 0) {
    autodebugRestore = debug;
  }
  debug = 1;
  autodebugExpires = jiffies + autodebug * HZ;
  autodebugcount --;
  printk(KERN_INFO "elmo: autodebug mode enabled\n");
}

static void elmo_autodebug_off(void) {
  if (autodebugExpires > 0 && autodebugExpires < jiffies) {
    debug = autodebugRestore;
    autodebugRestore = -1;
    autodebugExpires = 0;
    printk(KERN_INFO "elmo: autodebug mode expired\n");
  }
}

static void elmo_wait( int  mscount )
{
    int delay = ( mscount * HZ ) / 1000;

    wait_queue_head_t wait;

    init_waitqueue_head( &wait );

    if( delay <= 0 ) delay = 1;

    interruptible_sleep_on_timeout( &wait, delay );
}

static unsigned long
ReadRegister( unsigned long addr, unsigned width )
{
    unsigned long value;
    void  *vmem = (void *) (addr);

    switch( width ) {
        case 1:
            value = *((volatile unsigned char*)vmem);
            break;
        case 2:
            value = *((volatile unsigned short*)vmem);
            break;
        case 4:
            value = *((volatile unsigned long*)vmem);
            break;
        default:
            printk( KERN_ERR  "elmo: bad register read width\n" );
	    value = 0;
	    break;
    }

    return value;
}

static void
WriteRegister( unsigned long addr, unsigned value, unsigned width )
{
    void *vmem =(void *) (addr);

    switch( width ) {
        case 1:
            *((volatile unsigned char*)vmem) = value;
            break;
        case 2:
            *((volatile unsigned short*)vmem) = value;
            break;
        case 4:
            *((volatile unsigned long*)vmem) = value;
            break;
        default:
            printk( KERN_ERR  "elmo: bad register write width\n" );
    }
    barrier();
}


#ifdef NOTDEF
static void dump_bridge(void)
{
  int i;
  unsigned short value;
  printk(KERN_INFO "Bridge registers:");
  for (i = 0 ; i <= 32; i += 2) {
    value = *(volatile unsigned short *) KSEG1ADDR(ELMO_BRIDGE_BASE+i);
    printk(" %04X", value);
  }
  printk("\n");
  scsi_sleep(2);
}
#endif

static void __inline__ elmo_worker_next_state (struct elmo_worker *worker,
					 enum elmo_worker_state nextState) {
      worker->state = nextState;
      worker->go = 1;
      worker->dwell = 0;
}


/**
 *	elmo_init	-	initialise the Elmo adapter isntance
 *	@instance: adapter to configure
 *
 *	Initializes *instance
 *
 *	Notes : I assume that the host, hostno, and id bits have been
 *      set correctly.  I don't care about the irq and other fields. 
 *
 *	Locks: interrupts must be enabled when we are called 
 *
 *      XXX Pretty much all of this needs to be redone for Elmo!
 */

static void __init elmo_init(struct Scsi_Host *instance)
{
  int i;
  struct elmo_hostdata *hostdata = (struct elmo_hostdata *) instance->hostdata;
  struct elmo_worker *worker1, *worker2, *worker3;
  unsigned long flags;

  if(in_interrupt())
    printk(KERN_ERR "elmo_init called with interrupts off!\n");

  printk(KERN_INFO "elmo: initializing module\n");
  
  hostdata->aborted = 0;
  hostdata->pldLoaded = hostdata->atapiOK = 0;
  hostdata->id_mask = 1 << instance->this_id;
  for (i = hostdata->id_mask; i <= 0x80; i <<= 1)
    if (i > hostdata->id_mask)
      hostdata->id_higher_mask |= i;
  for (i = 0; i < ELMO_HANDLERS; ++i) {
    hostdata->busy[i] = 0;
    hostdata->pending[i] = 0;
    hostdata->connected[i] = NULL;
  }
  for (i = 0; i < 8; ++i) {
    hostdata->condition[i] = NothingPending;
  }
  hostdata->targets_present = 0;
  hostdata->issue_queue = NULL;
  
  if (!the_template) {
    the_template = instance->hostt;
    first_instance = instance;
  }

  hostdata->atapiSectorSize = 2048;
  hostdata->atapiBufferModulus = 2048;
  hostdata->time_expires = 0;
  hostdata->resourcesBusy = 0;
  hostdata->next_timer = NULL;
  hostdata->queueLock = RW_LOCK_UNLOCKED;
  INIT_LIST_HEAD(&hostdata->workers);

  hostdata->dmaDescriptors = kmalloc(PAGE_SIZE, GFP_KERNEL);

  printk(KERN_INFO "elmo: allocated DMA buffer at 0x%p, %ld bytes / %d = %ld descriptors\n",
	 hostdata->dmaDescriptors, PAGE_SIZE,
	 sizeof (struct BroadcomDMADescriptor),
	 PAGE_SIZE / sizeof (struct BroadcomDMADescriptor));

  hostdata->llReadBuffer = kmalloc(ELMO_READBUFFER_SIZE, GFP_KERNEL);
  hostdata->llReadBufferHead = hostdata->llReadBufferTail = 0;

  hostdata->llWriteBuffer = kmalloc(ELMO_READBUFFER_SIZE, GFP_KERNEL);
  hostdata->llWriteBufferHead = hostdata->llWriteBufferTail = 0;

  sema_init(&hostdata->writeSem, ELMO_MAX_PENDING_WRITERS);

  hostdata->bulkFifoSize = ELMO_BULK_FIFO_SIZE;

  thehostdata = hostdata; /* make it globally visible */

  /*
    Create and kick off a couple of workers to service the
    read and write queues
  */

  printk(KERN_INFO "elmo: initializing workers\n");

  worker1 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_ATOMIC);
  worker2 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_ATOMIC);
  worker3 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_ATOMIC);
  if (!worker1 || !worker2 || !worker3) {
    printk(KERN_ERR "elmo: could not create FIFO service worker\n");
    return;
  }
  memset(worker1, 0, sizeof *worker1);
  memset(worker2, 0, sizeof *worker2);
  memset(worker3, 0, sizeof *worker3);
  worker1->sn = ++thehostdata->workerSN;
  worker2->sn = ++thehostdata->workerSN;
  worker3->sn = ++thehostdata->workerSN;
  elmo_worker_next_state (worker1, ELMO_WORKER_SERVICE_READ_FIFO);
  elmo_worker_next_state (worker2, ELMO_WORKER_SERVICE_WRITE_FIFO);
  elmo_worker_next_state (worker3, ELMO_WORKER_AWAIT_RESET);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker1, &hostdata->workers);
  list_add_tail((struct list_head *) worker2, &hostdata->workers);
  list_add_tail((struct list_head *) worker3, &hostdata->workers);
  restore_flags(flags);
}

/*
  elmo_probe_bridge - see if the bridge PLD has been programmed.

  Check the bridge PLD to see if it's functional.  This is done by putting
  the Broadcom localbus into 16-bit 8-wait-state mode (so it won't
  hang if the PLD doesn't respond), writing a couple of scratch locations
  in the bridge and then reading them back.  If the values read back
  correctly, the bridge is assumed to be functional, and further ELMO
  probing and initialization can be initiated.

  Do NOT call this function after the ELMO has been initialized!  Doing
  so will probably yngvi up the bridge.

  Returns 1 if the bridge is alive, 0 if not.

*/

static int elmo_probe_bridge(void)
{

#define BCM_BGE_TEST_MODE __SYSREG(0xF0)

  unsigned short registerValue, revision, revisionInverted;

  printk(KERN_INFO "elmo: probing for bridge\n");

  scsi_sleep(1);

  // Setup the 7317 local bus so that Chip Select 3
  // is a 16-bit device with 8 clocks of waitstate.
  // This allows us to attempt to access the bridge's
  // revision registers, without hanging the system
  // if the bridge is non copos mentus.

  *BCM_EBI_CS3_CTRL = 0x0811FB11U;

  barrier();

  // Read the revision value twice (possible FPGA bug)

  registerValue = ELMO_BRIDGE->revision;
  barrier();
  registerValue = ELMO_BRIDGE->revision;
  barrier();
  revisionInverted = 0xFF & (registerValue >> 8);
  revision = 0xFF & registerValue;
  barrier();

  if ((revision ^ revisionInverted) == 0xFF && registerValue != 0xFF00) {
    printk(KERN_INFO "elmo: detected bridge, revision %02X, codeword %04X\n",
	   revision, registerValue);

    scsi_sleep(1);

    // Bridge is up and working OK.  Set CS3 to 16 bits with positive
    // acknowledgement required.

    *BCM_EBI_CS3_CTRL = 0x08004291U;
    WriteRegister(BCM_SPI_EBI_REQ_MUX,0,1);

    WriteRegister(BCM_GPIO_SEL1, ReadRegister( BCM_GPIO_SEL1, 1 ) | 0x43, 1 );
    /* 7317A0 has a bug in the receive, with hw flow control so, we use sw 
     * flow control. We could use hw flow control in the xmit case, but the
     * EBI dma engine cannot be disabled, so we cannot switch between 
     * enabling & disabling external DMA on EBI.
     */
    if (IS_7317A0()) {
        printk("elmo: detected 7317 A0\n");
        /* start EBI Tx, half-word, nonincr. sw flow control */
        *BCM_EBI_DMA_CTRL = 0x01000130; 
    }else {
        printk("elmo: detected 7317 A1\n");
        /* start EBI Tx, external DMA, word, nonincr. */
        *BCM_EBI_DMA_CTRL = 0xC1002030; 
    }

    *BCM_EBI_DMA_TX_START_ADDR_1 = ELMO_IDE_FIFO_ADDR;
    *BCM_EBI_DMA_TX_BUFFER_SIZE_1 = 0;

    *BCM_EBI_DMA_RX_START_ADDR_1 = ELMO_IDE_FIFO_ADDR;
    *BCM_EBI_DMA_RX_BUFFER_SIZE_1 = 0;

    *BCM_DMA_CH1_MAX_BURST = MAX_ISB_BURST/4; /* # words */
    *BCM_DMA_CH2_MAX_BURST = MAX_ISB_BURST/4; /* # words */ 

    *BCM_DMA_BYTE_REVERSE = 0x00000000;
    *BCM_DMA_HALF_REVERSE = 0x00000000;

    *BCM_DMA_CH1_CONFIG = 0x00000004; /* ISB DMA, chaining mode */
    *BCM_DMA_CH2_CONFIG = 0x00000004; /* ISB DMA, chaining mode */

    *BCM_DMA_CH1_INT_MASK = 4;
    *BCM_DMA_CH2_INT_MASK = 4;

    *(volatile unsigned short *) KSEG1ADDR(ELMO_DMA_EN) = 0;
    *(volatile unsigned short *) KSEG1ADDR(ELMO_IDE_WATERLINE_REG) = 0x4417U;
    *BCM_MTN_CLIENT_INDEX    = BCM_ISB_MMC_CLIENT_ID;
     barrier();
    *BCM_MTN_INDEX_DATA_WR   = 0x652D;
     barrier();

     barrier();
  } else {
    printk(KERN_INFO "elmo: bridge not detected (not yet initialized?) - read %04X\n",
	   registerValue);
    scsi_sleep(1);
    return -1;
  }

  // Enable interrupts from ELMO and bridge.  Don't enable CS8420 yet

  if (debug) printk(KERN_INFO "elmo: initializing ELMO bridge interrupts\n");

#ifndef SIMULATE_IRQ
  disable_irq(ELMO_IRQ);
#endif

  scsi_sleep(1);
  
  ELMO_BRIDGE->imr = 0;

  if (debug) printk(KERN_INFO "elmo: bridge initialization completed\n");
  scsi_sleep(1);

  // Kick the Elmo task into activity to start servicing the queues
  
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);

  return 0;

}

static void elmo_bridge_is_alive(struct Scsi_Host *instance)
{
  struct list_head *this, *next;
  struct elmo_worker *worker;
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  hostdata->pldLoaded = 1;
  notifier_installed = 1;
  ELMO_BRIDGE->stall = 0xFF; /* maximum timeout for stall */
  ELMO_BRIDGE->control = ELMO_CONTROL_LL_FIFO_RST | ELMO_CONTROL_IDE_FIFO_RST;
  ELMO_BRIDGE->control = ELMO_CONTROL_EBI_TIMEOUT_EN;
  ELMO_BRIDGE->imr = 
    ELMO_ISR_BUS_TIMEOUT    | ELMO_ISR_DMA_RD_IRQ     |
    ELMO_ISR_DMA_WR_IRQ     | ELMO_ISR_INT_FROM_ESS   | 
    ELMO_ISR_EBI_TO_ESS_OVR | ELMO_ISR_EBI_TO_ESS_UDR |
    ELMO_ISR_ESS_TO_EBI_OVR | ELMO_ISR_ESS_TO_EBI_UDR |
    ELMO_ISR_EBI_TO_IDE_OVR | ELMO_ISR_EBI_TO_IDE_UDR |
    ELMO_ISR_IDE_TO_EBI_OVR | ELMO_ISR_IDE_TO_EBI_UDR;
#ifdef SIMULATE_IRQ
  printk(KERN_INFO "elmo: starting polled simulation of IRQs\n");
  runIRQs = 1;
  queue_task(&elmo_timer_task_struct, &tq_timer);
#else
  enable_irq(ELMO_IRQ);
#endif
  list_for_each_safe(this,next,&hostdata->workers) {
    worker = (struct elmo_worker *) this;
    worker->go = 1;
  }
}

/*
 *	elmo_init_broadcom - set up the Broadcom CS3 chip-select address maps.
 */

static void elmo_init_broadcom (void)
{
  /* setup CS3 address space:
     base address TEI_BASE, LSBs specify size 1MB (7)
     8 bit device
     self terminating (sync = 0)
     8 wait states, 1 hold, 1 setup
  */
  *BCM_EBI_CS3_BASE = TEI_BASE + 7U;
  *BCM_EBI_CS3_CTRL = 0x0811FB01U;
}

/**
 * 	elmo_detect	-	look for Elmo controller
 *	@tpnt: the scsi template
 *
 *	Scan for the present of a Elmo ASIC.
 *
 *	Locks: none
 */

int __init elmo_detect(Scsi_Host_Template * tpnt)
{
  struct Scsi_Host *instance;
  int result;

  printk(KERN_INFO "elmo: version %s loaded\n", version);

  tpnt->proc_name = "elmo";

#ifndef SIMULATE_IRQ

  printk(KERN_INFO "elmo: trying to request bridge IRQ %d\n", ELMO_IRQ);
  
  if (request_irq(ELMO_IRQ, do_elmo_intr, SA_INTERRUPT, "Elmo", NULL)) {
    printk(KERN_INFO "elmo: failed to acquire IRQ %d\n", ELMO_IRQ);
    return 0;
  }

  printk(KERN_INFO "elmo: trying to request ISB DMA IRQ %d\n",
	 BCM_LINUX_DMA_IRQ);
  
  if (request_irq(BCM_LINUX_DMA_IRQ, broadcom_isb_intr, SA_INTERRUPT, "Elmo", NULL)) {
    printk(KERN_INFO "elmo: failed to acquire IRQ!\n");
    return 0;
  }
#endif

  printk(KERN_INFO "elmo: trying to register adapter\n");
  
  instance = scsi_register(tpnt, sizeof(struct elmo_hostdata));
  
  if (instance == NULL) {
    printk(KERN_INFO "elmo: failed to register adapter\n");
    free_irq(ELMO_IRQ, NULL);
    return 0;
  }
  
  printk(KERN_INFO "scsi%d : Elmo bridge at 0x%x irq %d", instance->host_no,
	 ELMO_BRIDGE_BASE, ELMO_IRQ);
  printk("\n");

  scsi_sleep(HZ / 4);

  elmo_init(instance);

  init_waitqueue_head( &readerWaitQueue );

  SET_MODULE_OWNER(&elmo_ll_fops);

  result = register_chrdev(ELMO_MAJOR, "elmo_ll", &elmo_ll_fops);
  if (result < 0) {
    printk(KERN_ERR "elmo_ll: can't get major %d\n",ELMO_MAJOR);
  }

  elmo_init_broadcom();

  printk(KERN_INFO "elmo: initializing tasks\n");
  elmo_task_struct.routine = elmo_task;
  elmo_task_struct.data = NULL;
  elmo_timer_task_struct.routine = elmo_timer_task;
  elmo_timer_task_struct.data = NULL;

  if (!elmo_probe_bridge()) {
    elmo_bridge_is_alive(instance);
  }

  return 1;
}

/**
 *	elmo_release_resources	-	free resources
 *	@instance: host adapter to clean up (There Can Be Only One)
 *
 *	Free the generic interface resources from this adapter.
 *
 *	Locks: none
 */
 
int elmo_release_resources(struct Scsi_Host *instance)
{
  free_irq(ELMO_IRQ, NULL);
  free_irq(DMA_IRQ, NULL);
  if (0 != unregister_chrdev(ELMO_MAJOR, "elmo_ll")) {
    printk(KERN_ERR "elmo_ll unregister failed\n");
  }
#ifdef SIMULATE_IRQ
  runIRQs = 0;
  scsi_sleep(HZ);
#endif
  return 0;
}

static void initialize_SCp(Scsi_Cmnd * cmd)
{
	/* 
	 * Initialize the Scsi Pointer field so that all of the commands in the 
	 * various queues are valid.
	 */

	if (cmd->use_sg) {
		cmd->SCp.buffer = (struct scatterlist *) cmd->buffer;
		cmd->SCp.buffers_residual = cmd->use_sg - 1;
		cmd->SCp.ptr = (char *) cmd->SCp.buffer->address;
		cmd->SCp.this_residual = cmd->SCp.buffer->length;
	} else {
		cmd->SCp.buffer = NULL;
		cmd->SCp.buffers_residual = 0;
		cmd->SCp.ptr = (char *) cmd->request_buffer;
		cmd->SCp.this_residual = cmd->request_bufflen;
	}

        if ( ! elmoCheckHostDMARules( cmd ) ) {

	  printk( "elmo driver: ERROR!!! Misaligned bulk data path pointers!\n" );

	}
}

/*
  Check the buffer(s) pointed to by a command, and see if all of them
  are compatible with the alignment and size rules for a Broadcom
  DMA transfer.  Every buffer must be 4-byte aligned, and a multiple of
  MAX_ISB_BURST bytes.
*/

static int elmoCheckHostDMARules (Scsi_Cmnd * cmd)
{
  int i;
  if (((long int) cmd->SCp.ptr) % 4 != 0 ||
      cmd->SCp.this_residual % MAX_ISB_BURST != 0) {
    return 0;
  }
  for (i = 1; i <= cmd->SCp.buffers_residual; i++) {
    if (((long int) cmd->SCp.buffer[i].address) % 4 != 0 ||
	cmd->SCp.buffer[i].length % MAX_ISB_BURST != 0) {
      return 0;
    }
  }
  return 1;
}

/*
  See if there is any more data to be transferred via DMA -
  return 0 if done, or residual size of next buffer if not done
*/

static int elmoCheckMoreHostDMA (Scsi_Cmnd * cmd)
{
  while (cmd->SCp.this_residual == 0 &&
	 cmd->SCp.buffers_residual > 0) {
    ++cmd->SCp.buffer;
    --cmd->SCp.buffers_residual;
    cmd->SCp.this_residual = cmd->SCp.buffer->length;
    cmd->SCp.ptr = cmd->SCp.buffer->address;
  }
  return cmd->SCp.this_residual;
}

/*
  Set up the Broadcom ROMbus DMA descriptor array.  Each descriptor
  is limited to [1,1023] words of word-aligned data transfer.
*/

#define ONEDMABURST 4032 /* 4096 - 64 (ISB burst size) */

static unsigned long elmoSetupHostDMA (struct Scsi_Host *instance,
				       Scsi_Cmnd * cmd,
				       unsigned long burstSize,
				       int direction,
				       unsigned int *descriptorCount)
{
  struct elmo_hostdata *hostdata;
  const int numDescriptors = PAGE_SIZE / sizeof (struct BroadcomDMADescriptor);
  int currentDescriptor;
  int waterline;
  int thisChunk, transferWordCount;
  unsigned long ioAddress;
  unsigned long consumed;
  hostdata = (struct elmo_hostdata *) (instance->hostdata);
  currentDescriptor = 0;
  waterline = 256 * 1024; /* size of bridge's PSR region */
  if (waterline > burstSize) {
    waterline = burstSize;
  }
  memset(hostdata->dmaDescriptors, 0, PAGE_SIZE);
  if (debug) printk(KERN_INFO "elmo: setting up Broadcom descriptors\n");
  ioAddress = ((unsigned long) ELMO_IDE_FIFO) & 0x3FFFFFFC;
  consumed = 0;
  while (waterline > 0 &&
	 currentDescriptor < numDescriptors &&
	 (thisChunk = elmoCheckMoreHostDMA(cmd)) > 0) {
    if (debug > 1) printk(KERN_INFO "elmo: want a chunk of %d bytes\n", thisChunk);
    if (thisChunk > waterline) {
      thisChunk = waterline;
      if (debug > 1) printk(KERN_INFO "elmo: clipped chunk to waterline: %d bytes\n", thisChunk);
    }
    if (thisChunk > ONEDMABURST) { /* XXX fix this - dplatt */
      thisChunk = ONEDMABURST;
      if (debug > 1) printk(KERN_INFO "elmo: clipped chunk to %d bytes\n", ONEDMABURST);
    }
    transferWordCount = thisChunk / 4;
    consumed += thisChunk;
    if (debug > 1) printk(KERN_INFO "elmo: want to transfer %d words from 0x%08lX\n",
			  transferWordCount, (unsigned long) cmd->SCp.ptr);
    hostdata->dmaDescriptors[currentDescriptor].transferByteCount = thisChunk;
    hostdata->dmaDescriptors[currentDescriptor].memoryAddr =
      (0x03FFFFFC & (unsigned int) cmd->SCp.ptr);
    hostdata->dmaDescriptors[currentDescriptor].f = (currentDescriptor == 0);
    hostdata->dmaDescriptors[currentDescriptor].ob = 1;
    if (debug > 1) printk(KERN_INFO "elmo: Broadcom DMA descriptor is %08lX %08lX\n",
			  *(unsigned long *) (&hostdata->dmaDescriptors[currentDescriptor]),
			  *(1 + (unsigned long *) (&hostdata->dmaDescriptors[currentDescriptor])));
    dma_cache_wback_inv((unsigned long) phys_to_virt((unsigned long) cmd->SCp.ptr), thisChunk);
    cmd->SCp.ptr           += thisChunk;
    cmd->SCp.this_residual -= thisChunk;
    waterline              -= thisChunk;
    currentDescriptor ++;
  }
  if (currentDescriptor == 0) {
    printk(KERN_ERR "elmo: AIEE!  Null ROMbus DMA!\n");
    panic("Null ROMbus DMA");
  }
  hostdata->dmaDescriptors[currentDescriptor - 1].l = 1;
  if (debug) printk(KERN_INFO "elmo: broadcom descriptors from %p to %p, consume %ld bytes\n",
		    &hostdata->dmaDescriptors[0],
		    &hostdata->dmaDescriptors[currentDescriptor - 1],
		    consumed);
  dma_cache_wback_inv((unsigned long) phys_to_virt((unsigned long) hostdata->dmaDescriptors), PAGE_SIZE);
  /*
    Flush the whole dcache.  Unfortunately, this seems to be necessary
    to ensure that SCSI Generic direct I/O works correctly.  The
    dma_cache_wback_inv() calls above *ought* to be sufficient, I
    think, and they work fine for non-direct I/O, but are not reliable
    for direct I/O.  I suspect that any "dirty" cache lines in
    user-virtual don't get flushed out correctly if direct I/O is
    used, because of the fact that the memory lines are multiply
    mapped (once for the user, and once for the kernel kiobuf).

    Ought to figure out a better way to fix this.  */
  flush_dcache();
  if (debug > 1) printk(KERN_INFO "elmo: set up %d Broadcom descriptors\n", currentDescriptor);
  *descriptorCount = currentDescriptor;
  return consumed;
}

/*
  Transfer data to the caller's buffer(s)
*/

static int elmoTransferToUser (Scsi_Cmnd * cmd, unsigned char *buffer, unsigned int length)
{
  int xferred = 0;
  int xfer;
  while (length > 0) {
    if (!cmd->SCp.this_residual) {
      if (cmd->SCp.buffers_residual) {
	++cmd->SCp.buffer;
	--cmd->SCp.buffers_residual;
	cmd->SCp.this_residual = cmd->SCp.buffer->length;
	cmd->SCp.ptr = cmd->SCp.buffer->address;
      } else {
	break;  /* ran out of space */
      }
    }
    xfer = length;
    if (xfer > cmd->SCp.this_residual) {
      xfer = cmd->SCp.this_residual;
    }
    memcpy(cmd->SCp.ptr, buffer, xfer);
    cmd->SCp.ptr += xfer;
    cmd->SCp.this_residual -= xfer;
    buffer += xfer;
    xferred += xfer;
    length -= xfer;
  }
  return xferred;
}

static int elmoTransferFromUser (Scsi_Cmnd * cmd, unsigned char *buffer, unsigned int length)
{
  int xferred = 0;
  int xfer;
  while (length > 0) {
    if (!cmd->SCp.this_residual) {
      if (cmd->SCp.buffers_residual) {
	++cmd->SCp.buffer;
	--cmd->SCp.buffers_residual;
	cmd->SCp.this_residual = cmd->SCp.buffer->length;
	cmd->SCp.ptr = cmd->SCp.buffer->address;
      } else {
	break;  /* ran out of space */
      }
    }
    xfer = length;
    if (xfer > cmd->SCp.this_residual) {
      xfer = cmd->SCp.this_residual;
    }
    memcpy(buffer, cmd->SCp.ptr, xfer);
    cmd->SCp.ptr += xfer;
    cmd->SCp.this_residual -= xfer;
    buffer += xfer;
    xferred += xfer;
    length -= xfer;
  }
  return xferred;
}

/*
  Fake up an INQUIRY response
*/

static int elmoInquiry (Scsi_Cmnd * cmd, int device,
		      int deviceType,
		      char *vendor,
		      char *model,
		      char *version)
{
  unsigned char inquiryData[36];
  int l;
  memset(inquiryData, 0, 16);
  memset(inquiryData+20, ' ', 20);
  inquiryData[0] = deviceType;
  inquiryData[3] = 2; /* response data format */
  inquiryData[4] = 36 - 4;
  l = strlen(vendor);
  if (l > 8) l = 8;
  strncpy(inquiryData+8, vendor, l);
  l = strlen(model);
  if (l > 16) l = 16;
  strncpy(inquiryData+16, model, l);
  l = strlen(version);
  if (l > 4) l = 4;
  strncpy(inquiryData+32, version, l);
  (void) elmoTransferToUser(cmd, inquiryData, 36);
  return 0;
}

/*
  Simulate a CHECK CONDITION error on an emulated device.
  XXX We need to cache the condition somewhere associated with the
  device so we can interpret it when we simulate REQUEST SENSE!
*/

static void elmoCheckCondition (struct Scsi_Host *instance,
			      Scsi_Cmnd * cmd, enum FakeCheckCondition cond)
{
  int device = cmd->target;
  printk(KERN_INFO "elmo: simulating Check Condition, code %d\n", cond);
  ((struct elmo_hostdata *) (instance->hostdata))->condition[device] = cond;
  cmd->SCp.Status = CHECK_CONDITION;
  cmd->result |= (CHECK_CONDITION << 1);
}

/*
  Fake up a Request Sense.  For now make it an
  Illegal Command Code in all cases.
*/

static void elmoRequestSense(struct Scsi_Host *instance,
			   Scsi_Cmnd * cmd)
{
  unsigned char senseData[8];
  int device = cmd->target;
  printk(KERN_INFO "elmo: simulating Request Sense, code %d\n", ((struct elmo_hostdata *) (instance->hostdata))->condition[device]);
  memset(senseData, 0, sizeof senseData);
  senseData[0] = 0x72; /* current error, descriptor sense data format */;
  switch (((struct elmo_hostdata *) (instance->hostdata))->condition[device]) {
  case NothingPending:
    senseData[1] = NO_SENSE;
    break;
  case UnitAttention:
    senseData[1] = UNIT_ATTENTION;
    senseData[2] = 0x28; /* NOT READY -> READY, MEDIUM MAY HAVE CHANGED */
    senseData[3] = 0x00; /* NOT READY -> READY, MEDIUM MAY HAVE CHANGED */
    break;
  case IllegalRequest:
    senseData[1] = ILLEGAL_REQUEST;
    senseData[2] = 0x20; /* INVALID COMMAND OPERATION CODE */
    senseData[3] = 0x00; /* INVALID COMMAND OPERATION CODE */
    break;
  case NotReady:
    senseData[1] = NOT_READY;
    senseData[2] = 0x67; /* CONFIGURATION FAILURE */
    senseData[3] = 0x00; /* CONFIGURATION FAILURE */
    break;
  case IllegalAddress:
    senseData[1] = ILLEGAL_REQUEST;
    senseData[2] = 0x24; /* INVALID FIELD IN CDB */
    senseData[3] = 0x00; /* INVALID FIELD IN CDB */
    break;
  default:
    senseData[1] = NOT_READY;
    senseData[2] = 0x20; /* ACCESS DENIED - NO ACCESS RIGHTS */
    senseData[3] = 0x02; /* ACCESS DENIED - NO ACCESS RIGHTS */
    break;
  }
  (void) elmoTransferToUser(cmd, senseData, 8);
  ((struct elmo_hostdata *) (instance->hostdata))->condition[device] =
    NothingPending;
}

static void elmo_bug(struct elmo_worker *worker, 
		     const char *reason)
{
  Scsi_Cmnd *cmd = (Scsi_Cmnd *) worker->scsiRequest;
  if (cmd) {
    /*
      Pass back a recognizable error to the caller
    */
    memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
    cmd->sense_buffer[0] = 0x72;
    cmd->sense_buffer[1] = HARDWARE_ERROR;
    cmd->sense_buffer[2] = 0x41; /* DATA PATH FAILURE */
    cmd->sense_buffer[3] = 0x00; /* DATA_PATH_FAILURE */
    cmd->result |= (DRIVER_SENSE << 24) | (CHECK_CONDITION << 1);
  }
  printk(KERN_ERR "elmo: I die.\n");
  printk(KERN_ERR "elmo:  %s.\n", reason);
  printk(KERN_ERR "elmo:   %s.\n", bug_line1[jiffies & 0x3]);
  printk(KERN_ERR "elmo:    %s.\n", bug_line2[(jiffies >> 2) & 0x3]);
  printk(KERN_ERR "elmo:     All die.  O the embarrassment!\n");
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_reset_ATA(struct elmo_hostdata *hostdata) {
  unsigned long flags;
  save_and_cli(flags);
  ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
  ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
  barrier();
  ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_RST_ACK;
  ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_RST_ACK;
  barrier();
  hostdata->wasReset = 0;
  printk(KERN_ERR "elmo: ATA FIFO cleared, ATA reset acknowledged\n");
  restore_flags(flags);
}

static int elmoBridgeDispatchFunction (struct Scsi_Host *instance,
				     Scsi_Cmnd * cmd, int device) 
{
  struct elmo_worker *worker;
  struct elmo_hostdata *hostdata = (struct elmo_hostdata *) (instance->hostdata);
  unsigned long flags;
  (void) device;
  if (debug) printk(KERN_INFO "elmo: bridge dispatch function for cmd %p\n", cmd);
  switch (cmd->cmnd[0]) {
  case TEST_UNIT_READY:
    cmd->result = 0; /* it's OK */;
    return 1;
  case INQUIRY:
    (void) elmoInquiry(cmd, device, TYPE_COMM, "TiVo", "Elmo bridge", "0001");
    cmd->result = 0; /* it's OK */;
    return 1;
  case REQUEST_SENSE:
    elmoRequestSense(instance, cmd);
    cmd->result = 0; /* it's OK */;
    return 1;
  case READ_10:
  case WRITE_10:
    break;
  default:
    elmoCheckCondition(instance, cmd, IllegalRequest);
    return 1;
  }
  (void) worker;
  (void) flags;
  if (!hostdata->pldLoaded) {
    elmoCheckCondition(instance, cmd, NotReady);
    return 1;
  }
  /*
    Create a Worker for this request and queue it up...
  */
  if (debug > 1) printk(KERN_INFO "elmo: allocating worker\n");
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
				       GFP_ATOMIC);
  if (debug > 1) printk(KERN_INFO "elmo: got worker at %p\n", worker);
  if (!worker) {
    return -1; /* requeue it for later */
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++hostdata->workerSN;
  worker->scsiRequest = cmd;
  if (debug) printk(KERN_INFO "elmo: worker is #%d\n", worker->sn);
  elmo_worker_next_state(worker, ELMO_WORKER_SETUP);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker, &hostdata->workers);
  /*
    Take it out of the Connected slot and mark us non-busy
  */
  hostdata->connected[ELMO_HANDLER_BRIDGE] = NULL;
  hostdata->busy[ELMO_HANDLER_BRIDGE] = 0;
  restore_flags(flags);
  return 0;
}

static int elmoBridgeIoctlFunction (struct Scsi_Host *instance,
				  Scsi_Device *dev, int cmd, void *arg)
{
  int result = 0;
  unsigned long flags;
  struct elmo_hostdata *hostdata = (struct elmo_hostdata *) (instance->hostdata);

  (void) dev;

  if (cmd != kElmoDownloadPLD && cmd != kElmoSetDebugMode &&
      !hostdata->pldLoaded) {
    printk(KERN_INFO "elmo: access ioctl() on a dead bridge!\n");
    return -EINVAL;
  }
    
  switch( cmd )
    {
    case kElmoReadBridge:
      {
	ElmoIODescriptor descriptor;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{
	  short *dest = (short *)(descriptor.data);
	  volatile short *src = GetELMOBridgeAddress( descriptor.address );
	  int i;

	  for( i = 0; i < descriptor.length; i += 2 )
	    {
	      short value = *src;
	      barrier();
	      put_user( value, dest );
	      if (debug > 1) {
		printk(KERN_INFO "elmo: peek bridge %08lX has  %04X\n",
		       (unsigned long) src, value);
	      }
	      ++src; ++dest;
	    }
	}
	break;
      }
    case kElmoWriteBridge:
      {
	ElmoIODescriptor descriptor;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{
	  short *src = (short *)(descriptor.data);
	  volatile short *dest = GetELMOBridgeAddress( descriptor.address );
	  int i;

	  for( i = 0; i < descriptor.length; i += 2 )
	    {
	      short value;
	      get_user( value, src );
	      if (debug > 1) {
		printk(KERN_INFO "elmo: poke bridge %08lX with %04X\n",
		       (unsigned long) dest, value);
	      }
	      *dest = value;
	      barrier();
	      ++src; ++dest;
	    }
	}
	break;
      }

    case kElmoWriteBridgeMasked:
      {
	ElmoMaskDescriptor descriptor;
	volatile short *dest;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );
	dest = GetELMOBridgeAddress( descriptor.address );

	save_and_cli(flags);

	*dest = (*dest & ~descriptor.mask) |
	  (descriptor.bits & descriptor.mask);

	restore_flags(flags);

	break;
      }

    case kElmoDownloadPLD:
      {
	ElmoPLDDescriptor descriptor;
	int i;
	
	if (((struct elmo_hostdata *) (instance->hostdata))->pldLoaded) {
	  printk(KERN_INFO "elmo: WARNING - reprogramming a live bridge!\n");
	  ((struct elmo_hostdata *) (instance->hostdata))->pldLoaded = 0;
#ifdef SIMULATE_IRQ
	  runIRQs = 0;
#endif
	  scsi_sleep(HZ);
	}

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{

	  unsigned char *src = (unsigned char *)(descriptor.data);

	  volatile unsigned char *dest = (unsigned char *) ELMO_BRIDGE;

	  // PROGRAM is connected to GPIO 33
	  // WRITE signal is connected to GPIO 3
        
	  WriteRegister( BCM_GPIO_SEL0, 
			 ReadRegister( BCM_GPIO_SEL0, 1 ) & ~0x30U,
			 1 );

	  // Set both signals off (GPIO 1)
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2,
			 1 );

	  // output rather than open drain output
	  WriteRegister( BCM_GPIO_BASE+8, 
			 ReadRegister( BCM_GPIO_BASE+8, 1 ) & ~0x8U,
			 1 );
	  // output rather than input
	  WriteRegister( BCM_GPIO_BASE+7, 
			 ReadRegister( BCM_GPIO_BASE+7, 1 ) & ~0x8U,
			 1 );
                       
	  // output rather than open drain output
	  WriteRegister( BCM_GPIO_BASE+0x55, 
			 ReadRegister( BCM_GPIO_BASE+0x55, 1 ) & ~0x2U,
			 1 );
	  // output rather than input
	  WriteRegister( BCM_GPIO_BASE+0x53, 
			 ReadRegister( BCM_GPIO_BASE+0x53, 1 ) & ~0x2U,
			 1 );

	  // Set both signals off (GPIO 1)
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2,
			 1 );

	  // Pulse the "Program" signal to the FPGA low and then high. 
        
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) & ~0x2U,
			 1 );
	  elmo_wait(1);
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2U,
			 1 );

	  // Set the "Write" signal to the FPGA low. 
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) & ~0x8U,
			 1 );

	  printk(KERN_INFO "elmo: loading %ld bytes from 0x%p\n",
		 descriptor.length, src);

	  for( i = 0; i < descriptor.length; ++i )
	    {
	      unsigned char value;
	      get_user( value, src );
	      *dest = value;
	      ++src;
	    }

	  elmo_wait ( 1 );

	  /* take FPGA out of programming mode */
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );

	  /* make CS3 be a 16 bit device again */
          *BCM_EBI_CS3_CTRL = 0x08004291U;
          *(volatile unsigned short *) KSEG1ADDR(ELMO_IDE_WATERLINE_REG) = 0x4417U;
          *BCM_MTN_CLIENT_INDEX    = BCM_ISB_MMC_CLIENT_ID;
          barrier();
          *BCM_MTN_INDEX_DATA_WR   = 0x652D;
	  barrier();

	  // Probe the bridge to see if it responds.

	  if (!elmo_probe_bridge()) {
	    elmo_bridge_is_alive(instance);
	  }

	}
	break;
      }
    case kElmoSetDebugMode:
      {
	debug = (int) arg;
	printk(KERN_INFO "elmo: debug level set to %d\n", debug);
	break;
      }
    case kElmoReserve:
      {
	break; /* not yet implemented, treat as a NO-OP */
      }
    case kElmoRelease:
      {
	break; /* not yet implemented, treat as a NO-OP */
      }
    case kElmoReportDiagMode:
      {
	hostdata->diagMode = (int) arg;
	break;
      }
    case kElmoResetESS:
      {
	printk(KERN_INFO "elmo: rude reset of ESS\n");

	save_and_cli(flags);

	ELMO_BRIDGE->control |= ELMO_CONTROL_ESS_RST;
	barrier();
	ELMO_BRIDGE->control &= ~ELMO_CONTROL_ESS_RST;

	restore_flags(flags);
	
	break;
      }
    default:
      {
	result = -EINVAL;
      }
    }
  return result;
}


static void elmo_release_chip_resources (struct Scsi_Host *instance, unsigned int resourceMask)
{
  struct list_head *this;
  struct elmo_worker *worker;
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  unsigned int releaseMask;
  unsigned int intersectMask;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  releaseMask = resourceMask;
  hostdata->resourcesBusy &= ~resourceMask;
  if (debug > 1) printk(KERN_INFO "elmo: releasing resource mask 0x%08X\n",
			resourceMask);
  save_and_cli(flags);
  list_for_each(this,&hostdata->workers) {
    worker = (struct elmo_worker *) this;
    if (worker->elmoResourcesUsed & resourceMask) {
      worker->elmoResourcesUsed &= ~resourceMask;
    }
    intersectMask = worker->elmoResourcesNeeded & releaseMask;
    if (intersectMask) {
      if (debug > 1) printk(KERN_INFO "elmo: worker %d wants 0x%08X\n",
			    worker->sn, intersectMask);
      worker->elmoResourcesNeeded &= ~intersectMask;
      worker->go = 1;
      /*      intersectMask &= ~ELMO_MULTIPLE_GRABBERS; */
      releaseMask &= ~intersectMask;
    }
  }
  restore_flags(flags);
}

/*
  Other exported routines, for other drivers which must manipulate
  the ELMO bridge registers.
*/

int ELMOReadBridge(unsigned long address, unsigned long length, void *data )
{
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  unsigned long *dataWord;
  if (!first_instance) {
    return -ENODEV;
  }
  hostdata = (struct elmo_hostdata *) first_instance->hostdata;
  if (!hostdata->pldLoaded) {
    return -ENODEV;
  }
  dataWord = (unsigned long *) data;
  save_and_cli(flags);
  while (length > 0) {
    *dataWord = *((unsigned long *) (ELMO_BRIDGE_BASE + address));
    length -= sizeof (unsigned long);
    address += sizeof (unsigned long);
    dataWord += sizeof (unsigned long);
  }
  restore_flags(flags);
  return 0;
}

int ELMOWriteBridge(unsigned long address, unsigned long length, void *data )
{
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  unsigned long *dataWord;
  if (!first_instance) {
    return -ENODEV;
  }
  hostdata = (struct elmo_hostdata *) first_instance->hostdata;
  if (!hostdata->pldLoaded) {
    return -ENODEV;
  }
  dataWord = (unsigned long *) data;
  save_and_cli(flags);
  while (length > 0) {
    *((unsigned long *) (ELMO_BRIDGE_BASE + address)) = *dataWord;
    length -= sizeof (unsigned long);
    address += sizeof (unsigned long);
    dataWord += sizeof (unsigned long);
  }
  restore_flags(flags);
  return 0;
}

int ELMOWriteBridgeMasked(unsigned long address, unsigned short mask,
			unsigned short bits)
{
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  if (!first_instance) {
    return -ENODEV;
  }
  hostdata = (struct elmo_hostdata *) first_instance->hostdata;
  if (!hostdata->pldLoaded) {
    return -ENODEV;
  }
  save_and_cli(flags);
  *((unsigned long *) (ELMO_BRIDGE_BASE + address)) =
    (~mask & *((unsigned long *) (ELMO_BRIDGE_BASE + address))) |
    ( mask & bits);
  restore_flags(flags);
  return 0;
}

#ifdef NOTDEF

static void show_pci_intr_status(void)
{
  printk(KERN_INFO "elmo: PCI interrupt enable %08lX status %08lX\n",
	 *BCM_PCI_INTR_REG, *BCM_PCI_INTR_STAT);
}

#endif

  
/*
  State-machine nodes for the bulk-data transfer channel.
*/

static void elmo_worker_setup(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  unsigned long int length, i;
  Scsi_Cmnd *cmd;  
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  cmd = (Scsi_Cmnd *) worker->scsiRequest;
  if (cmd->timeout_per_command) {
    worker->potentialDeadman = cmd->timeout_per_command; /* convert to ticks */
  }
  memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
  length = cmd->SCp.this_residual;
  if (debug > 1) printk(KERN_INFO "elmo: buffer %p(%ld)\n", cmd->SCp.ptr, length);
  for (i = 1; i <= cmd->SCp.buffers_residual; i++) {
    if (debug > 1) printk(KERN_INFO "elmo: %p(%d)\n", cmd->SCp.buffer[i].address,
		      cmd->SCp.buffer[i].length);
    length += cmd->SCp.buffer[i].length;
  }
  if (debug) printk(KERN_INFO "elmo: setup request, command 0x%02X for %ld bytes\n", cmd->cmnd[0],
	 length);
  worker->sc_data_direction = cmd->sc_data_direction;
  worker->transferSize = length;
  worker->transferRemaining = length;
  worker->residual = length;
  elmo_worker_next_state(worker, ELMO_WORKER_GET_BULK_INTERFACE);
  switch (cmd->cmnd[0]) {
  case READ_10:
    if (worker->sc_data_direction != SCSI_DATA_READ) {
      elmo_bug(worker, "The READ_10 command had wrong sc_data_direction");
    }
    break;
  case WRITE_10:
    if (worker->sc_data_direction != SCSI_DATA_WRITE) {
      elmo_bug(worker, "The WRITE_10 command had wrong sc_data_direction");
    }
    break;
  default:
    elmo_bug(worker, "An unknown SCSI command got into setup");
    break;
  }
}

static void elmo_worker_done(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  Scsi_Cmnd *cmd;  
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  int device, handler;
  if (debug) printk(KERN_INFO "elmo: worker %d is done\n", worker->sn);
  if (worker->dataBuffer) {
    kfree(worker->dataBuffer);
    worker->dataBuffer = NULL;
  }
  if (worker->holdingSemaphore) {
    up(&hostdata->writeSem);
    if (debug > 1) printk(KERN_INFO "elmo: worker %d up sem to %d\n",
			  worker->sn,
			  sem_getcount(&hostdata->writeSem));
  }
  if (worker->signalATA) {
    /*
      XXX This isn't really right - we'd like to generate an ATA-bus
      "done" event by writing to an ATA-bus register, and perhaps even
      pass along error information if there's a problem.  Alas, the
      FPGA doesn't provide us with access to the necessary registers.
    */
    ELMO_BRIDGE->control |= ELMO_CONTROL_INT_ESS;
    barrier();
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_INT_ESS;
  }
  if (hostdata->wasReset) {
    /*
      Previous command might have left the ATA in a funky state.
      If so, clear the FIFOs out and acknowledge the reset, and
      continue onwards.
    */
    if (debug > 1) printk(KERN_INFO "elmo: worker %d clearing prior reset\n",
			  worker->sn);
    elmo_reset_ATA(hostdata);
  }
  elmo_release_chip_resources(instance, worker->elmoResourcesUsed);
  cmd = (Scsi_Cmnd *) worker->scsiRequest;
  if (cmd) {
    if (worker->reportHang) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting a data path failure\n");
      memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
      cmd->sense_buffer[0] = 0x72;
      cmd->sense_buffer[1] = HARDWARE_ERROR;
      cmd->sense_buffer[2] = 0x41; /* DATA PATH FAILURE */
      cmd->sense_buffer[3] = 0x00; /* DATA_PATH_FAILURE */
      cmd->result |= (DRIVER_SENSE << 24) | (DID_ERROR << 16) | (CHECK_CONDITION << 1);
      cmd->retries = cmd->allowed;
    } else if (worker->reportError) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting an error\n");
      cmd->result |= DID_ERROR << 16;
    } else if (worker->retries) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting retries\n");
      cmd->result |= (DRIVER_SENSE << 24) | (CHECK_CONDITION << 1);
      cmd->retries = cmd->allowed;
    } else if (worker->aborting) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting a command abort\n");
      cmd->result |= DID_ABORT << 16;
    }
    cmd->resid = worker->residual;
    if (debug) printk(KERN_INFO "elmo: cmd result is %08X\n", cmd->result);
    device = cmd->target;
    handler = deviceDispatcher[device];
    if (hostdata->connected[handler] == cmd) {
      hostdata->connected[handler] = NULL;
      hostdata->busy[handler] = 0;
    }
    cmd->scsi_done(cmd);
  }
  elmo_autodebug_off();  /* turn off autodebug if it has timed out */
}

static void elmo_worker_get_bulk_interface(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (worker->aborting) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (!(hostdata->resourcesBusy & ELMO_RESOURCE_BULK_INTERFACE)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_BULK_INTERFACE;
    hostdata->resourcesBusy |= ELMO_RESOURCE_BULK_INTERFACE;
    if (debug > 1) printk(KERN_INFO "elmo: worker %d has bulk I/O\n",
			  worker->sn);
    /*
      Now that we own it, make sure we're notified if the ESS
      tries a reset while we're using it
    */
    worker->elmoInterruptsFired &= ~ ELMO_INT_DSK_RST_REQ;
    worker->elmoInterruptsPending |= ELMO_INT_DSK_RST_REQ;
    if (hostdata->wasReset) {
      /*
	Previous command might have left the ATA in a funky state.
	If so, clear the FIFOs out and acknowledge the reset, and
	continue onwards.
      */
      if (debug > 1) printk(KERN_INFO "elmo: worker %d clearing prior reset\n",
			    worker->sn);
      elmo_reset_ATA(hostdata);
    }
    elmo_worker_next_state(worker, ELMO_WORKER_GET_ESS_COMMAND);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_BULK_INTERFACE;
    if (debug > 1) printk(KERN_INFO "elmo: worker %d queued for bulk I/O\n",
			  worker->sn);
  }
}

static void elmo_worker_get_ess_command(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  Scsi_Cmnd *cmd = (Scsi_Cmnd *) worker->scsiRequest;
  unsigned char commandByte;
  unsigned int sectorCountByte;
  unsigned short lba;
  unsigned long flags;
  if (worker->aborting) {
    worker->reportHang = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  /*
    If the ESS reset the bus while we were in charge of it, we've
    got bad problems.  Reset the bus and then bail out.
  */
  if (hostdata->wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while awaiting command\n", worker->sn);
    elmo_reset_ATA(hostdata);
    worker->reportHang = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  /*
    If the ESS got its command to us before we were really ready for
    it, the disk-command-received interrupt will be trying to fire.
    Save a bit of time by acknowledging and clearing it here, and 
    pretending that it had fired.
  */
  save_and_cli(flags);
  if (!hostdata->haveCommand) {
    /*
      No command yet, wait for it
    */
    if (debug > 1) printk(KERN_INFO "elmo: worker %d queued for ATA command\n",
			  worker->sn);
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_ATA_COMMAND;
    restore_flags(flags);
    return;
  }
  /*
    Command has arrived.  Capture the registers.
  */
  hostdata->haveCommand = 0;
  worker->ataCommandReg = hostdata->ataCommandReg;
  worker->ataCylinderReg = hostdata->ataCylinderReg;
  worker->signalATA = 1;
  restore_flags(flags);
  if (debug > 1) printk(KERN_INFO "elmo: worker %d has ATA command\n",
			worker->sn);
  commandByte = worker->ataCommandReg & 0xFF;
  sectorCountByte = (worker->ataCommandReg >> 8) & 0xFF;
  if (sectorCountByte == 0 && ESS_ATA_TRANSFER_ABORT != commandByte) {
    sectorCountByte = 256;
  }
  if (debug > 1) printk(KERN_INFO "elmo: worker %d has ATA command %02X for %d sectors\n",
			worker->sn, commandByte, sectorCountByte);
  switch (commandByte) {
  case WIN_READ:
  case WIN_READDMA:
  case WIN_MULTREAD:
    if (cmd->cmnd[0] != WRITE_10) {
      elmo_bug(worker, "The ESS asks to read, but we are not writing");
      worker->errorATA = 1;
      return;
    }
    break;
  case WIN_WRITE:
  case WIN_WRITEDMA:
  case WIN_MULTWRITE:
    if (cmd->cmnd[0] != READ_10) {
      elmo_bug(worker, "The ESS asks to write, but we are not reading");
      worker->errorATA = 1;
      return;
    }
    break;
  case ESS_ATA_TRANSFER_ABORT:
    {
      Scsi_Cmnd *cmd = (Scsi_Cmnd *) worker->scsiRequest;
      if (debug > 1) printk(KERN_INFO "elmo: received vendor specific abort from ESS\n");
      if (cmd) {
        /*
          Pass back a recognizable error to the caller
        */
        memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
        cmd->sense_buffer[0] = ESS_ATA_TRANSFER_ABORT;
        cmd->sense_buffer[1] = sectorCountByte;
        cmd->sense_buffer[2] = 0x41; /* DATA PATH FAILURE */
        cmd->sense_buffer[3] = 0x00; /* DATA_PATH_FAILURE */
        cmd->sense_buffer[4] = worker->ataCylinderReg >> 8;
        cmd->sense_buffer[5] = worker->ataCylinderReg & 0xFFU;
        cmd->result |= (DRIVER_SENSE << 24) | (CHECK_CONDITION << 1);
      }
      if (debug > 1) printk(KERN_INFO "elmo: ESS says abort, so aborting bulk data transfer\n");
      worker->aborting = 1;
      elmo_worker_next_state(worker, ELMO_WORKER_DONE);
      ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_CMD_ACK;
      ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_CMD_ACK;
      return;
    }
  default:
    elmo_bug(worker, "The ESS sends a command we do not understand");
    worker->errorATA = 1;
    return;
  }
  lba = (cmd->cmnd[4] << 8) + cmd->cmnd[5];
  if (lba != worker->ataCylinderReg) {
    printk( KERN_ERR
	    "elmo: TiVo expecting ATA transfer ID %d, ESS asking for %d\n",
	    lba, worker->ataCylinderReg );
    elmo_bug(worker, "The command sequence numbers do not match");
    worker->errorATA = 1;
    return;
  } else if (sectorCountByte * 512 > worker->transferRemaining) {
    printk( KERN_ERR
	    "elmo: TiVo ready to transfer %d bytes, ESS wants %d bytes\n", 
	    worker->transferRemaining, sectorCountByte * 512 );
    elmo_bug(worker, "The ESS wants to transfer too much data");
    worker->errorATA = 1;
    return;
  } 
  worker->transferBurst = sectorCountByte * 512;
  worker->transferRemaining -= worker->transferBurst;
  /*
      We've got the command and are ready to acknowledge it so the
      ESS can start the data transfer.  If they're doing a write,
      we can't start our DMA immediately - it would stall on a PY - so
      we queue up for the data-in-FIFO interrupt and go to sleep.
      If they're doing a read, we can go set up our DMA now.
    */
  if (commandByte == WIN_READ || 
      commandByte == WIN_READDMA || 
      commandByte == WIN_MULTREAD) {
    if (debug > 1) printk(KERN_INFO "elmo: can go push data to FIFO\n");
#ifdef TEMPORARY_SLOW_DMA_HACK
    elmo_queue_worker_for_timer(worker, 10, ELMO_WORKER_SETUP_DMA);
#else
    elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
#endif
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: must wait for data in FIFO\n");
    elmo_queue_worker_for_interrupt (worker, ELMO_INT_IDE_TO_EBI_NE, 
				     ELMO_WORKER_AWAIT_FIFO);
    ELMO_BRIDGE->fmr |= ELMO_FIFO_IDE_TO_EBI_NE;
  }
}

static void elmo_worker_await_fifo(struct Scsi_Host *instance,
				   struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (hostdata->wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while awaiting FIFO\n",
	    worker->sn);
    elmo_reset_ATA(hostdata);
    worker->aborting = 1;
  }
  if (worker->aborting) {
    worker->reportHang = 1;
    worker->errorATA = 1;
    ELMO_BRIDGE->fmr &= ~ELMO_FIFO_IDE_TO_EBI_NE;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (worker->elmoInterruptsFired & ELMO_INT_IDE_TO_EBI_NE) {
    ELMO_BRIDGE->fmr &= ~ELMO_FIFO_IDE_TO_EBI_NE;
#ifdef TEMPORARY_SLOW_DMA_HACK
      elmo_queue_worker_for_timer(worker, 10, ELMO_WORKER_SETUP_DMA);
#else
      elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
#endif
  } else {
    printk(KERN_INFO "elmo: worker awaiting IDE FIFO woke for no reason\n");
  }
}

static void elmo_worker_setup_dma(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  Scsi_Cmnd *cmd = (Scsi_Cmnd *) worker->scsiRequest;
  unsigned long consumed;
  unsigned long flags;
  unsigned short watermarkBit;
  unsigned long watermarkInterrupt;
  unsigned int direction;
  unsigned int transferBurst;
  unsigned int descriptorCount;
  unsigned int transferCount = 0;
  unsigned int myDMA = 0;

  if (worker->sc_data_direction == SCSI_DATA_READ) {
    direction = 0;
    watermarkBit = ELMO_FIFO_IDE_TO_EBI_HWM;
    watermarkInterrupt = ELMO_INT_IDE_TO_EBI_HWM;
  } else {
    direction = 1;
    watermarkBit = ELMO_FIFO_EBI_TO_IDE_LWM;
    watermarkInterrupt = ELMO_INT_EBI_TO_IDE_LWM;
  }
  save_and_cli(flags);
  ELMO_BRIDGE->fmr &= ~watermarkBit;
  worker->elmoInterruptsFired &= ~ watermarkInterrupt;
  worker->elmoInterruptsPending &= ~ watermarkInterrupt;
  restore_flags(flags);
  if (hostdata->wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while prepping DMA\n",
	    worker->sn);
    elmo_reset_ATA(hostdata);
    worker->aborting = 1;
  }
  if (worker->aborting) {
    worker->reportHang = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (worker->transferBurst == 0) {
    if (worker->residual == 0) {
      if (debug) printk(KERN_INFO "elmo: all data has been transferred\n");
      elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    } else {
      if (debug) printk(KERN_INFO "elmo: more data to transfer\n");
      elmo_worker_next_state(worker, ELMO_WORKER_GET_ESS_COMMAND);
    }
    return;
  }
  save_and_cli(flags);
  if (!(ELMO_BRIDGE->fsr & watermarkBit)) {
    elmo_queue_worker_for_interrupt (worker, watermarkInterrupt,
				     ELMO_WORKER_SETUP_DMA);
    ELMO_BRIDGE->fmr |= watermarkBit;
    if (debug) printk(KERN_INFO "elmo: awaiting FIFO watermark\n");
    restore_flags(flags);
    return;
  } 
  restore_flags(flags);
  transferBurst = worker->transferBurst;
  if (hostdata->bulkFifoSize != 0 &&
      transferBurst > hostdata->bulkFifoSize / 2) {
    transferBurst = hostdata->bulkFifoSize / 2;
  }

  /* 7317A0 has a bug in the receive, with hw flow control so, we use sw flow
   * control. We could use hw flow control in the xmit case, but the EBI dma
   * engine cannot be disabled, so we cannot switch between enabling &
   * disabling external DMA on EBI.
   */
  if (!IS_7317A0()) {
      transferBurst = (1024*128); /* max 128k transfers */
  }

  consumed = elmoSetupHostDMA (instance, cmd, transferBurst, direction,
			       &descriptorCount);
  worker->consuming = consumed;
  worker->transferBurst -= consumed;
  transferCount = consumed/(MAX_ISB_BURST); 
  /*
    Set up the Broadcom DMA registers and start the engine.
  */
  save_and_cli(flags);
  if (direction) {
    if (debug) printk(KERN_INFO "elmo: starting Broadcom write-DMA burst\n");

    *BCM_DMA_CH2_BUFF_ADDR = 
      (((unsigned long) hostdata->dmaDescriptors) & 0x03FFFFFF);
    *BCM_DMA_CH2_BCNTR_LEN = descriptorCount;

    if (IS_7317A0()) {
        *BCM_EBI_DMA_TX_BUFFER_SIZE_1 = consumed;
    }

    if (0 != swaponwrite) {
        *(volatile unsigned short *) KSEG1ADDR(ELMO_DMA_EN) = 0x8000U; 
    }else {
        *(volatile unsigned short *) KSEG1ADDR(ELMO_DMA_EN) = 0; 
    }
    barrier();


    elmo_broadcom_isb_enabled = 1;

    worker->dmaRunning = ELMO_DMA_WRITING;

    if (!IS_7317A0()) {
        myDMA = ELMO_INT_DMA_WR_IRQ;
        elmo_asic_write_dma_enabled = 1;
    }else {
        myDMA = 0;
        elmo_asic_write_dma_enabled = 0;
    }


    *BCM_EBI_DMA_CTRL |= 0x00000032U; /* start EBI Tx, word, nonincr. */
    barrier();
    /* enable write dma flow control */
    if (!IS_7317A0()) {
       *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)|=(0x2000U|(transferCount&0xFFFU)); 
       barrier();
    }
    *BCM_DMA_CH2_CONFIG |= 0x00000001U; /* start ISB DMA, chaining mode */
    barrier();
  } else {

    if (debug) printk(KERN_INFO "elmo: starting Broadcom read-DMA burst\n");
    *BCM_DMA_CH1_BUFF_ADDR =
      (((unsigned long) hostdata->dmaDescriptors) & 0x03FFFFFF);
    *BCM_DMA_CH1_BCNTR_LEN = descriptorCount;

    if (IS_7317A0()) {
        *BCM_EBI_DMA_RX_BUFFER_SIZE_1 = consumed;
    }
    if (0 != swaponread) {
        *(volatile unsigned short *) KSEG1ADDR(ELMO_DMA_EN) = 0x4000U; 
    }else {
        *(volatile unsigned short *) KSEG1ADDR(ELMO_DMA_EN) = 0;
    }

    barrier();

    *BCM_DMA_CH1_CONFIG |= 0x00000001U; /* start ISB DMA, chaining mode */
    *BCM_EBI_DMA_CTRL   |= 0x00000031U; /* start EBI Rx, word, nonincr. */
    barrier();

    elmo_broadcom_isb_enabled = 1;

    worker->dmaRunning = ELMO_DMA_READING;

    if (!IS_7317A0()) {
        myDMA = ELMO_INT_DMA_RD_IRQ;
        elmo_asic_read_dma_enabled = 1;
    }else {
        myDMA = 0;
        elmo_asic_read_dma_enabled = 0;
    }

    /* enable read dma flow control */
    if (!IS_7317A0()) {
       *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)|=(0x1000U|(transferCount&0xFFFU)); 
       barrier();
    }
  }

  elmo_queue_worker_for_interrupt (worker, myDMA |
                                   ELMO_INT_ISB_DMA|
				   ELMO_INT_BUS_TIMEOUT |
				   ELMO_INT_EBI_TO_ESS_OVR |
				   ELMO_INT_ESS_TO_EBI_UDR,
				   ELMO_WORKER_FINISH_DMA);
  enable_irq(BCM_LINUX_DMA_IRQ);
  restore_flags(flags);

#if 0
#define FAKE_ATA_RESET_FROM_ESS
#ifdef FAKE_ATA_RESET_FROM_ESS
  if (!direction) {
    ELMO_BRIDGE->control |=  ELMO_CONTROL_ATA_INT_RST;
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_ATA_INT_RST;
    barrier();
  }
#endif /* FAKE_ATA_RESET_FROM_ESS */
#endif /* 0 */
}

static void elmo_stop_isb_dma( volatile u32* isb_config_reg)
{
    int count = 10000;

    if (!elmo_broadcom_isb_enabled) {
        return;
    }

    /* stop ISB DMA */
    do {
        barrier();

        *isb_config_reg |= 0x00000002U; /* stall ISB DMA */

        barrier();
        udelay(1000);
    }while((!(*isb_config_reg & 0x00000002U)) && count--);

    if (!count) {
        printk("elmo : unable to stall isb dma\n");
        panic("elmo : unable to stall isb dma\n");
    }

    barrier();

    *isb_config_reg = 0x00000040U;
    barrier();

    count = 10000;

    do {
        barrier();
        *isb_config_reg &= 0xFFFFFFFEU; /* stop ISB DMA */
        barrier();
        udelay(1000);
    }while((*isb_config_reg & 0x00000001U) && count--);

    if (!count) {
        printk("elmo : unable to stop isb dma\n");
        panic("elmo : unable to stop isb dma\n");
    }

    *isb_config_reg = 0x00000004U;
    elmo_broadcom_isb_enabled = 0;
}

static void elmo_worker_finish_dma(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  unsigned int resetATA = 0;
  unsigned int finished = 0;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  save_and_cli(flags);
  if (hostdata->wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
    printk(KERN_ERR "elmo: worker %d hit by ATA reset while doing DMA\n",
	   worker->sn);
    resetATA = 1;
    worker->aborting = 1;
    worker->reportHang = 1;
  }
  if (worker->elmoInterruptsFired & ELMO_INT_BUS_TIMEOUT) {
    printk(KERN_ERR "elmo: Oh bother.  Worker %d stalled the Broadcom DMA\n",
		      worker->sn);
    worker->aborting = 1;
    worker->reportHang = 1;
  }
  if (worker->elmoInterruptsFired & (ELMO_INT_EBI_TO_ESS_OVR |
				     ELMO_INT_ESS_TO_EBI_UDR)) {
    printk(KERN_ERR "elmo: Oh fudge.  Worker %d thrashed the FIFO\n",
	   worker->sn);
    worker->aborting = 1;
    worker->reportHang = 1;
  }
  if (worker->aborting) {
    switch (worker->dmaRunning) {
    case ELMO_DMA_IDLE:
      break;
    case ELMO_DMA_READING:
        if (elmo_broadcom_isb_enabled) {
            printk(KERN_INFO "elmo: aborting DMA read\n");

            if (elmo_asic_read_dma_enabled) {
                /* set the FPGA DMA transfer count to 0 but let it run.
                 * This will ensure that the current burst is transferred and
                 * we will not have data stuck in the any of the FIFOs. Stop
                 * the FPGA DMA after ISB has been stopped
                 */
                *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0xFFFU);
                barrier();
            }

            elmo_stop_isb_dma(BCM_DMA_CH1_CONFIG);

            /* stop EBI RX DMA. 
             * This will only be effective on a 7317-A0 system
             */
            *BCM_EBI_DMA_CTRL   &= ~0x00000001U; 
            barrier();
            *BCM_EBI_DMA_CTRL   |= 0x00000030U; /* clear EBI interrupt flags */

            if (elmo_asic_read_dma_enabled) {
                /* stop the FPGA DMA */
                *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0x1000U);
                barrier();
                elmo_asic_read_dma_enabled = 0;
            }

            ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
            ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
            barrier();

      }
      worker->dmaRunning = ELMO_DMA_IDLE;
      break;
    case ELMO_DMA_WRITING:
      if ( elmo_asic_write_dma_enabled || 
           ( IS_7317A0() && elmo_broadcom_isb_enabled )) {
	printk(KERN_INFO "elmo: aborting DMA write\n");

        /* stop ISB DMA */
        elmo_stop_isb_dma(BCM_DMA_CH2_CONFIG);

        /* reset the FPGA FIFO to make room for any pending ISB DMA to 
         * go through 
         */
        ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
        barrier();

        /* set the FPGA DMA transfer count to 0 but let it run.
         * This will ensure that the current burst is transferred and
         * we will not have data stuck in the any of the FIFOs. Stop
         * the FPGA DMA after ISB has been stopped
         */
        *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0xFFFU);
        barrier();

        /* stop EBI TX DMA. 
         * This will only be effective on a 7317-A0 system
         */
        *BCM_EBI_DMA_CTRL   &= ~0x00000002U;
        barrier();
        *BCM_EBI_DMA_CTRL   |= 0x00000030U; /* clear EBI interrupt flags */

        /* stop the FPGA DMA */
        *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0x2000U);
        barrier();
        /* reset the FPGA FIFOs again */
        ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
        barrier();
	elmo_asic_write_dma_enabled = 0;
      }
      worker->dmaRunning = ELMO_DMA_IDLE;
      break;
    }
    if (resetATA) {
      elmo_reset_ATA(hostdata);
    }
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    restore_flags(flags);
    return;
  }
  switch (worker->dmaRunning) {
  case ELMO_DMA_IDLE:
  default:
    printk(KERN_INFO "elmo: worker %d lost state of Broadcom DMA\n",
	   worker->sn);
    worker->aborting = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    break;
  case ELMO_DMA_READING:
    if (!elmo_asic_read_dma_enabled && !elmo_broadcom_isb_enabled) {
      finished = 1;
    }
    if (debug) printk(KERN_INFO "elmo: worker %d finds read DMAs %sdone\n",
		      worker->sn, finished ? "" : "not ");
    break;
  case ELMO_DMA_WRITING:
    if (!elmo_asic_write_dma_enabled && !elmo_broadcom_isb_enabled) {
      finished = 1;
    }
    if (debug) printk(KERN_INFO "elmo: worker %d finds write DMAs %sdone\n",
		      worker->sn, finished ? "" : "not ");
      break;
  }
  if (finished) {
    if (debug) printk(KERN_INFO "elmo: worker %d ends Broadcom DMA\n",
		      worker->sn);
    worker->residual -= worker->consuming;
    elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
  }
  restore_flags(flags);
}

static void elmo_worker_await_reset(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  unsigned long flags, fired;
  int i;
  save_and_cli(flags);
  fired = worker->elmoInterruptsFired;
  worker->elmoInterruptsFired = 0;
  restore_flags(flags);
  if (fired & ELMO_INT_DSK_RST_REQ) {
    if (debug > 1) printk(KERN_INFO "elmo: shmoo sees that ATA reset has arrived\n");
    hostdata->haveCommand = 0; /* previous command voided! */
    if (hostdata->resourcesBusy & ELMO_RESOURCE_BULK_INTERFACE) {
      /*
	Another worker is in charge of the bus.  Make sure that this
	worker sees the reset, and acknowledges it, before releasing
	the bus.
      */
      hostdata->wasReset = 1;
    } else {
      /*
	No one has the bus at the moment.  Acknowledge the reset now
	so that the AT-host can unblock.
      */
      elmo_reset_ATA(hostdata);
    }
  }
  if (fired & ELMO_INT_DSK_CMD_REQ) {
    if (debug > 1) printk(KERN_INFO "elmo: shmoo sees that ATA command has arrived\n");
    if (hostdata->haveCommand) {
      printk(KERN_ERR "elmo: command arrived with command pending!\n");
    }
    hostdata->haveCommand = 1;
    hostdata->ataCommandReg = ELMO_BRIDGE->command;
    hostdata->ataCylinderReg = ELMO_BRIDGE->cylinder;
    /*
      Acknowledge the command, turn the ESS loose.
    */
    ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_CMD_ACK;
    barrier();
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_CMD_ACK;
    switch (hostdata->ataCommandReg & 0xFF) {
    case WIN_READ:
    case WIN_READDMA:
    case WIN_WRITE:
    case WIN_WRITEDMA:
    case WIN_MULTWRITE:
    case WIN_MULTREAD:
      elmo_release_chip_resources(instance, ELMO_RESOURCE_ATA_COMMAND);
      break;
    case WIN_IDENTIFY:
      printk(KERN_INFO "elmo: performing dummy IDENTIFY\n");
      save_and_cli(flags);
      for (i = 0; i < 32; i++) {
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x4E75;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x6D62;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x6572;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x206E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x696E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x652E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x2E2E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x2020;
      }
      hostdata->haveCommand = 0;
      restore_flags(flags);
      break;
  case ESS_ATA_TRANSFER_ABORT:
    {
      struct list_head *this, *next;
      struct elmo_worker *bulk_worker = NULL;

      list_for_each_safe(this,next,&hostdata->workers) {
        struct elmo_worker *tmp_worker = (struct elmo_worker *) this;
        if (tmp_worker->elmoResourcesUsed & ELMO_RESOURCE_BULK_INTERFACE) {
            bulk_worker = tmp_worker;
            break;
        }
      }

      if (NULL == bulk_worker) {
          if (debug > 1) printk(KERN_INFO "elmo: received vendor specific abort from ESS when there was no worker waiting on bulk IO.\n");
          elmo_release_chip_resources(instance, ELMO_RESOURCE_ATA_COMMAND);
          break;
      }

      if (debug > 1) printk(KERN_INFO "elmo: received vendor specific abort from ESS\n");

      save_and_cli(flags);
      hostdata->haveCommand = 0;
      bulk_worker->aborting = 1;
      bulk_worker->go = 1;
      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);
      Scsi_Cmnd *cmd = (Scsi_Cmnd *) bulk_worker->scsiRequest;
      unsigned int sectorCountByte = (bulk_worker->ataCommandReg >> 8) & 0xFF;
      restore_flags(flags);

      if (cmd) {
        /*
          Pass back a recognizable error to the caller
        */
        memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
        cmd->sense_buffer[0] = ESS_ATA_TRANSFER_ABORT;
        cmd->sense_buffer[1] = sectorCountByte;
        cmd->sense_buffer[2] = 0x41; /* DATA PATH FAILURE */
        cmd->sense_buffer[3] = 0x00; /* DATA_PATH_FAILURE */
        cmd->sense_buffer[4] = bulk_worker->ataCylinderReg >> 8;
        cmd->sense_buffer[5] = bulk_worker->ataCylinderReg & 0xFFU;
        cmd->result |= (DRIVER_SENSE << 24) | (CHECK_CONDITION << 1);
      }
      if (debug > 1) printk(KERN_INFO "elmo: ESS says abort, so aborting bulk data transfer\n");
      break;
    }
    default:
      printk(KERN_ERR "elmo: unhandled command 0x%02X!\n",
	     hostdata->ataCommandReg & 0xFF);
      save_and_cli(flags);
      ELMO_BRIDGE->error = (ABRT_ERR << 8);
      ELMO_BRIDGE->error = (ABRT_ERR << 8) | 1;
      ELMO_BRIDGE->error = 0;
      hostdata->haveCommand = 0;
      restore_flags(flags);
    }
  }
  elmo_queue_worker_for_interrupt(worker, 
				  ELMO_INT_DSK_RST_REQ | ELMO_INT_DSK_CMD_REQ,
				  ELMO_WORKER_AWAIT_RESET);
  ELMO_BRIDGE->imr |= ELMO_ISR_DSK_RST_REQ | ELMO_ISR_DSK_CMD_REQ;
}

static void elmo_worker_setup_reset(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (debug) printk(KERN_INFO "elmo: FIFO-flush DMA initiated, wait 5 ms\n");
  elmo_queue_worker_for_timer(worker, 5, ELMO_WORKER_RESET_DONE);
}

static void elmo_worker_reset_done(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (debug) printk(KERN_INFO "elmo: timer expired, stopping DMA\n");
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_worker_service_read_fifo(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  int spaceRemaining;
  int took;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!hostdata->pldLoaded) {
    if (debug > 1) printk(KERN_INFO "elmo: not ready, pull worker sleeps\n");
    return;
  }
  elmo_worker_disarm_timer(worker);
  spaceRemaining = (hostdata->llReadBufferTail - hostdata->llReadBufferHead) %
    ELMO_READBUFFER_SIZE;
  if (spaceRemaining == 1) {
    /* the ring buffer is actually full.  Go away until someone drains it */
    if (debug > 1) printk(KERN_INFO "elmo: pull buffer is full, worker sleeps\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READBUFFER_SPACE;
    return;
  }
  if (spaceRemaining == 0) {
    spaceRemaining = ELMO_READBUFFER_SIZE;
  }
  if (debug > 1) printk(KERN_INFO "elmo: pull buffer has %d bytes available\n",
			spaceRemaining);
  took = 0;
  while (spaceRemaining > 1 && (ELMO_BRIDGE->fsr & ELMO_FIFO_ESS_TO_EBI_NE)) {
    hostdata->llReadBuffer[hostdata->llReadBufferHead] =
      (unsigned char) *(ELMO_LOW_LATENCY_FIFO);
    hostdata->llReadBufferHead = (hostdata->llReadBufferHead + 1) %
      ELMO_READBUFFER_SIZE;
    took ++;
    spaceRemaining --;
  }
  PLOG(
      PLOG_ELMO_RPC_NEW_DATA_FROM_ESS,
      sizeof(took),
      &took
      );
  PLOG(
      PLOG_ELMO_RPC_NEW_DATA_WAITING_FOR,
      sizeof(spaceRemaining),
      &spaceRemaining
      );

  if (debug > 1) printk(KERN_INFO "elmo: pulled %d bytes to buffer\n",
			took);

#if 0
  /*
    If we didn't get anything this time, arrange to be interrupted
    instantly whenever anything appears in the FIFO.

    Otherwise, arrange to be interrupted if the FIFO reaches the
    partially-full watermark, or after a few ticks.  Also, release the
    data-is-available resource so that the reader (if any) can awaken.
  */
  if (took == 0) {
    if (debug > 1) printk(KERN_INFO "elmo: pullworker sleeps awaiting data\n");
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_NE,
				    ELMO_WORKER_SERVICE_READ_FIFO);
    ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_HWM) |
      ELMO_FIFO_ESS_TO_EBI_NE;
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: pull worker sleeps on watermark\n");
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_HWM,
				    ELMO_WORKER_SERVICE_READ_FIFO);
    elmo_queue_worker_for_timer(worker, 5, ELMO_WORKER_SERVICE_READ_FIFO);
    ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_NE) |
      ELMO_FIFO_ESS_TO_EBI_HWM;
    elmo_release_chip_resources(instance, ELMO_RESOURCE_READBUFFER_DATA);
  }

#else

  /*
    The logic above makes things very slooow. Scanning directory takes 
    38ms per directory entry, for example.

    Instead, we will wait for NE interrupt always, even if we copy some data.
    It seems the above logic is there to protect us from too much data and 
    too many interrupts. Trouble with that logic is that we could read some,
    and then wait for HWM or the 5 timer ticks, and if the HWM never hits, 
    we suffer 5 timer ticks.
  */
  if (debug > 1) printk(KERN_INFO "elmo: pullworker sleeps awaiting data\n");
  elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_NE,
                                  ELMO_WORKER_SERVICE_READ_FIFO);
  ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_HWM) |
      ELMO_FIFO_ESS_TO_EBI_NE;
  if ( took > 0 ) {
      elmo_release_chip_resources(instance, ELMO_RESOURCE_READBUFFER_DATA);
  }

#endif

}

static void elmo_worker_service_write_fifo(struct Scsi_Host *instance, struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  int dataAvailable;
  int took;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!hostdata->pldLoaded) {
    if (debug > 1) printk(KERN_INFO "elmo: not ready, push worker sleeps\n");
    return;
  }
  dataAvailable = (hostdata->llWriteBufferTail - hostdata->llWriteBufferHead) %
    ELMO_WRITEBUFFER_SIZE;
  if (dataAvailable == 0) {
    if (debug > 1) printk(KERN_INFO "elmo: push buffer is empty, worker sleeps\n");
    /* the ring buffer is empty.  Go away until someone fills it*/
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_DATA;
    return;
  }
  if (debug > 1) printk(KERN_INFO "elmo: push buffer has %d bytes in it\n",
			dataAvailable);
  took = 0;
  while (dataAvailable > 0 && !(ELMO_BRIDGE->fsr & ELMO_FIFO_EBI_TO_ESS_FF)) {
    *(ELMO_LOW_LATENCY_FIFO) = 
      (unsigned short) hostdata->llWriteBuffer[hostdata->llWriteBufferHead];
    hostdata->llWriteBufferHead = (hostdata->llWriteBufferHead + 1) %
      ELMO_WRITEBUFFER_SIZE;
    dataAvailable --;
    took ++;
  }

  PLOG( PLOG_ELMO_RPC_DATA_SENT_TO_FIFO,
        sizeof(dataAvailable),
        &dataAvailable
      );

  if (debug > 1) printk(KERN_INFO "elmo: pushed %d bytes to FIFO\n", took);
  /*
    If there's any data remaining to be sent, arrange to be interrupted
    when the amount of data in the FIFO falls below the watermark.

    Otherwise, arrange to be awakened when somebody puts more data in
    our buffer... and, also, send the "data space is available" signal
    now to awaken any sleeping writers.
  */
  if (dataAvailable > 0) {
    if (debug > 1) printk(KERN_INFO "elmo: push worker sleeps until FIFO below waterline\n");
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_EBI_TO_ESS_LWM,
				    ELMO_WORKER_SERVICE_WRITE_FIFO);
    ELMO_BRIDGE->fmr = ELMO_BRIDGE->fmr | ELMO_FIFO_EBI_TO_ESS_LWM;
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: push worker sleeps until data appears\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_DATA;
    elmo_release_chip_resources(instance, ELMO_RESOURCE_WRITEBUFFER_SPACE);
  }
}

static void elmo_worker_gate_writer(struct Scsi_Host *instance,
				    struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!(hostdata->resourcesBusy & ELMO_RESOURCE_WRITER_GATE)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_WRITER_GATE;
    hostdata->resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITER_GATE;
  }
}

static void elmo_worker_become_writer(struct Scsi_Host *instance,
				      struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!(hostdata->resourcesBusy & ELMO_RESOURCE_WRITER)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_WRITER;
    hostdata->resourcesBusy |= ELMO_RESOURCE_WRITER;
    elmo_worker_next_state(worker, ELMO_WORKER_WRITER);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITER;
  }
}

static void elmo_worker_writer(struct Scsi_Host *instance,
			       struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  int dataAvailable, spaceAvailable;
  int i;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  dataAvailable = (hostdata->llWriteBufferTail - hostdata->llWriteBufferHead) %
    ELMO_WRITEBUFFER_SIZE;
  spaceAvailable = ELMO_WRITEBUFFER_SIZE - dataAvailable - 1;
  if (spaceAvailable < worker->dataBufferSize) {
    /* Insufficient space.  Go away until the buffer is drained*/
    if (debug > 1) printk(KERN_INFO "elmo: push buffer full, worker sleeps\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_SPACE;
    return;
  }
  for (i = 0; i < worker->dataBufferSize; i++) {
    hostdata->llWriteBuffer[hostdata->llWriteBufferTail] = worker->dataBuffer[i];
    hostdata->llWriteBufferTail = (hostdata->llWriteBufferTail + 1) %
      ELMO_WRITEBUFFER_SIZE;
  }
  if (debug > 1) printk(KERN_INFO "elmo: pushed %d bytes to ring buffer\n", i);
  /*
    Kick the write servicer
  */
  elmo_release_chip_resources(instance, ELMO_RESOURCE_WRITEBUFFER_DATA);
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_worker_become_reader(struct Scsi_Host *instance,
				      struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!(hostdata->resourcesBusy & ELMO_RESOURCE_READER)) {
    if (debug > 1) printk(KERN_INFO "elmo: worker %d becomes the reader\n",
			  worker->sn);
    worker->elmoResourcesUsed |= ELMO_RESOURCE_READER;
    hostdata->resourcesBusy |= ELMO_RESOURCE_READER;
    elmo_worker_next_state(worker, ELMO_WORKER_READER);
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: worker %d waiting to become the reader\n",
			  worker->sn);
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READER;
  }
}

static void elmo_worker_reader(struct Scsi_Host *instance,
			       struct elmo_worker *worker)
{
  struct elmo_hostdata *hostdata;
  int dataAvailable;
  int i;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  dataAvailable = (hostdata->llReadBufferHead - hostdata->llReadBufferTail) %
    ELMO_WRITEBUFFER_SIZE;
  if (dataAvailable == 0) {
    /* No data.  Go away until the read servicer provides some */
    if (debug > 1) printk(KERN_INFO "elmo: reader sleeping to await data\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READBUFFER_DATA;
    return;
  }
  if (debug > 1) printk(KERN_INFO "elmo: reader finds %d avail, %d wanted\n",
			dataAvailable, worker->dataBufferSize);
  for (i = 0; dataAvailable > 0 && i < worker->dataBufferSize; i++) {
    worker->dataBuffer[i] = hostdata->llReadBuffer[hostdata->llReadBufferTail];
    hostdata->llReadBufferTail = (hostdata->llReadBufferTail + 1) %
      ELMO_WRITEBUFFER_SIZE;
    dataAvailable --;
    worker->dataBufferCount ++;
  }
  /*
    Wake up the process which submitted us
  */
  worker->readCompleted = 1;
  wake_up_interruptible(&readerWaitQueue);
  /*
    Kick the read servicer
  */
  elmo_release_chip_resources(instance, ELMO_RESOURCE_READBUFFER_SPACE);
  /*
    Go into the post-read idle state until the process which submitted
    us grabs the data and marks us as being done.
  */
  elmo_worker_next_state(worker, ELMO_WORKER_READER_POST);
}

static void elmo_worker_reader_post(struct Scsi_Host *instance,
				    struct elmo_worker *worker)
{
  /* This is a null/idling state.  Do nothing, go back to sleep. */
  (void) instance;
  if (debug > 1) printk(KERN_INFO "elmo: worker %d awaiting postprocessing\n",
			worker->sn);
}

static void elmo_queue_worker_for_interrupt (struct elmo_worker *worker,
					  unsigned long interruptsNeeded,
					  enum elmo_worker_state nextState)
{
  unsigned long flags;
  save_and_cli(flags);
  if (debug > 1 && worker->elmoInterruptsPending & ~interruptsNeeded) {
    printk(KERN_DEBUG "elmo: worker %d new state %d old int %08lX add int %08lX\n",
	   worker->sn, nextState, worker->elmoInterruptsPending,
	   interruptsNeeded);
  }
  worker->elmoInterruptsPending |= interruptsNeeded;
  worker->state = nextState;
  restore_flags(flags);
}

static void elmo_queue_worker_for_timer (struct elmo_worker *worker,
				       int timerTicks,
				       enum elmo_worker_state nextState)
{
  unsigned long flags;
  save_and_cli(flags);
  worker->elmoInterruptsPending |= ELMO_INT_TIMER;
  worker->timeout = timerTicks ? timerTicks : 1;
  worker->state = nextState;
  restore_flags(flags);
  queue_task(&elmo_timer_task_struct, &tq_timer);
}

static void elmo_worker_disarm_timer (struct elmo_worker *worker) {
  unsigned long flags;
  save_and_cli(flags);
  worker->elmoInterruptsPending &= ~ELMO_INT_TIMER;
  worker->timeout = 0;
  restore_flags(flags);
}

static void elmo_worker_set_deadman (struct elmo_worker *worker,
					 int timerTicks)
{
  unsigned long flags;
  save_and_cli(flags);
  worker->deadman = timerTicks ? timerTicks : 20*HZ;
  restore_flags(flags);
  queue_task(&elmo_timer_task_struct, &tq_timer);
}

static void elmo_worker_disarm_deadman (struct elmo_worker *worker)
{
  worker->deadman = 0;
}

static unsigned long elmo_interrupt_for_workers(struct Scsi_Host *instance,
					      unsigned long interruptsFired)
{
  unsigned long flags;
  unsigned long intersection;
  unsigned long interruptsTaken;
  struct list_head *this;
  struct elmo_worker *worker;
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  save_and_cli(flags);
  interruptsTaken = 0;
  if (debug) printk(KERN_INFO "elmo: process virtual interrupt 0x%08lX\n",
	 interruptsFired);
  list_for_each(this,&hostdata->workers) {
    worker = (struct elmo_worker *) this;
    if (worker->elmoInterruptsPending) {
      if (debug) printk(KERN_INFO "elmo: worker %d pending on 0x%08lX\n",
			worker->sn, worker->elmoInterruptsPending);
      intersection = worker->elmoInterruptsPending & interruptsFired;
      if (intersection) {
	interruptsTaken |= intersection;
	worker->elmoInterruptsFired = intersection;
	worker->elmoInterruptsPending &= ~intersection;
	worker->go = 1;
	elmo_worker_disarm_timer(worker);
	if (debug) printk(KERN_INFO "elmo: woke up the worker\n");
      }
    }
  }
  return interruptsTaken;
}  

/*
 *	elmo_run_workers - run the individual state machines for the
 *	ATAPI requests on the queue.
 *
 */

static int elmo_run_workers(struct Scsi_Host *instance)
{
  int done = 1;
  struct list_head *this, *next;
  struct elmo_worker *worker;
  struct elmo_hostdata *hostdata;
  hostdata = (struct elmo_hostdata *) instance->hostdata;
  if (!hostdata->pldLoaded) {
    return done;
  }
  if (debug > 2) printk(KERN_INFO "elmo: running worker queue\n");
  list_for_each_safe(this,next,&hostdata->workers) {
    worker = (struct elmo_worker *) this;
    if (worker->go) {
      if (debug > 2) printk(KERN_INFO "elmo: worker %d go!\n", worker->sn);
      worker->go = 0;

      PLOG( PLOG_ELMO_RPC_WORKER_STATE, sizeof(worker->state), &worker->state );

      switch (worker->state) {
      case ELMO_WORKER_SETUP:
	elmo_worker_setup(instance, worker);
	break;
      case ELMO_WORKER_GET_BULK_INTERFACE:
	elmo_worker_get_bulk_interface(instance, worker);
	break;
      case ELMO_WORKER_GET_ESS_COMMAND:
	elmo_worker_get_ess_command(instance, worker);
	break;
      case ELMO_WORKER_SETUP_DMA:
	elmo_worker_setup_dma(instance, worker);
	break;
      case ELMO_WORKER_AWAIT_FIFO:
	elmo_worker_await_fifo(instance, worker);
	break;
      case ELMO_WORKER_FINISH_DMA:
	elmo_worker_finish_dma(instance, worker);
	break;
      case ELMO_WORKER_DONE:
	elmo_worker_done(instance, worker);
	list_del((struct list_head *) worker);
	kfree(worker);
	worker = NULL;
	break;
      case ELMO_WORKER_AWAIT_RESET:
	elmo_worker_await_reset(instance, worker);
	break;
      case ELMO_WORKER_SETUP_RESET:
	elmo_worker_setup_reset(instance, worker);
	break;
      case ELMO_WORKER_RESET_DONE:
	elmo_worker_reset_done(instance, worker);
	break;
      case ELMO_WORKER_SERVICE_READ_FIFO:
	elmo_worker_service_read_fifo(instance, worker);
	break;
      case ELMO_WORKER_SERVICE_WRITE_FIFO:
	elmo_worker_service_write_fifo(instance, worker);
	break;
      case ELMO_WORKER_GATE_WRITER:
	elmo_worker_gate_writer(instance, worker);
	break;
      case ELMO_WORKER_BECOME_WRITER:
	elmo_worker_become_writer(instance, worker);
	break;
      case ELMO_WORKER_WRITER:
	elmo_worker_writer(instance, worker);
	break;
      case ELMO_WORKER_BECOME_READER:
	elmo_worker_become_reader(instance, worker);
	break;
      case ELMO_WORKER_READER:
	elmo_worker_reader(instance, worker);
	break;
      case ELMO_WORKER_READER_POST:
	elmo_worker_reader_post(instance, worker);
	break;
      default:
	break;
      }
    } else {
      if (debug > 2) printk(KERN_INFO "elmo: worker %d nogo\n", worker->sn);
    }
  }
  list_for_each(this,&hostdata->workers) {
    worker = (struct elmo_worker *) this;
    done &= ! worker->go;
  }
  if (debug > 2) printk(KERN_INFO "elmo: queue runner says %sdone\n",
			done ? "" : "not ");
  return done;
}

/*
  The coroutine code has been gently and lovingly lifted from
  NCR5380 - go see the comments there.
*/

static unsigned long main_running = 0;

static __inline__ void run_main(void)
{
  if (!test_and_set_bit(0, &main_running)) {
    elmo_main();
  }
}

/*
  Coroutine runner task
*/

static void elmo_task (void *whatever)
{
  (void) whatever;
  run_main();
}

/*
  Timer task.  Run through all known workers, and kick awake any of
  them who were waiting for a timer countdown and whose countdown
  has expired.  Also, abort any workers which had a deadman timer
  (these are usually background/buffered write commands, whose
  SCSI request/command has already been returned as 'completed'
  and which thus no longer have a SCSI-layer timer watching over
  them.
*/

static void elmo_timer_task (void *whatever)
{
   struct Scsi_Host *instance;
   struct elmo_hostdata *hostdata;
   struct list_head *this;
   struct elmo_worker *worker;
   int wakeup = 0;
   int rearm = 0;
   (void) whatever;
   for (instance = first_instance;
	instance && instance->hostt == the_template;
	instance = instance->next) {
     hostdata = (struct elmo_hostdata *) instance->hostdata;
     list_for_each(this,&hostdata->workers) {
       worker = (struct elmo_worker *) this;
       if (worker->deadman != 0) {
	 worker->deadman --;
	 if (worker->deadman == 0) {
	   worker->aborting = 1;
	   worker->go = 1;
	   printk(KERN_ERR "elmo: deadman timer aborting worker #%d\n", worker->sn);
	 } else {
	   rearm = 1;
	 }
       }
       if (worker->timeout > 0 &&
	   (worker->elmoInterruptsPending & ELMO_INT_TIMER)) {
	 worker->timeout --;
	 if (worker->timeout == 0) {
	   worker->elmoInterruptsPending &= ~ELMO_INT_TIMER;
	   worker->elmoInterruptsFired |= ELMO_INT_TIMER;
	   worker->go = 1;
	   wakeup = 1;
	 } else {
	   rearm = 1;
	 }
       }
     }
   }
   if (wakeup) {
     queue_task (&elmo_task_struct, &tq_immediate);
     mark_bh (IMMEDIATE_BH);
   }
#ifdef SIMULATE_IRQ
   do_elmo_intr(ELMO_IRQ, NULL, NULL);
   rearm = runIRQs;
#endif
   if (rearm) {
     queue_task(&elmo_timer_task_struct, &tq_timer);
   }
}


/**
 *	elmo_main	-	Elmo state machines
 *
 *	elmo_main is a coroutine that runs as long as more work can 
 *      be done on the Elmo host adapter in a system.  Both 
 *      elmo_queue_command() and elmo_intr() will try to start it 
 *      in case it is not running.
 * 
 *	Locks; The caller must hold the io_request_lock. The lock will still be
 *	held on return but may be dropped while running. Called functions take
 *	the DMA lock.
 */

static void elmo_main(void) {
   Scsi_Cmnd *tmp, *prev, *next;
   struct Scsi_Host *instance;
   struct elmo_hostdata *hostdata;
   int done;
   unsigned int iflags, handler;

   /*
    * We run (with interrupts disabled) until we're sure that none of
    * the host adapters have anything that can be done, at which point 
    * we set main_running to 0 and exit.
    *
    * Interrupts are enabled before doing various other internal 
    * instructions, after we've decided that we need to run through
    * the loop again.
    *
    * this should prevent any race conditions.
    *
    * XXX the Elmo adaptation of this process would not be
    * safe if two or more ELMO adapters were present, and one adapter
    * were unloaded while the main loop was running... unlikely but
    * possible.
    */

   do {
     /* Lock held here */
     done = 1;
     for (instance = first_instance; instance && instance->hostt == the_template; instance = instance->next) {
       /*
	* Search through the issue_queue for a command destined
	* for a target that's not busy.
	*/
       hostdata = (struct elmo_hostdata *) instance->hostdata;
       write_lock_irqsave(&hostdata->queueLock, iflags);
       for (tmp = (Scsi_Cmnd *) hostdata->issue_queue, prev = NULL;
	    tmp;
	    prev = tmp, tmp = next) 
	 {
	   int device = tmp->target;
	   next = (Scsi_Cmnd *) tmp->host_scribble;
	   /*  When we find one, remove it from the issue queue. */
	   if (device >= ELMO_DEVICES || !(hostdata->busy[deviceDispatcher[device]])) {
	     if (prev) {
	       REMOVE(prev, prev->host_scribble, tmp, tmp->host_scribble);
	       prev->host_scribble = tmp->host_scribble;
	     } else {
	       REMOVE(-1, hostdata->issue_queue, tmp, tmp->host_scribble);
	       hostdata->issue_queue = (Scsi_Cmnd *) tmp->host_scribble;
	     }
	     tmp->host_scribble = NULL;
	     if (device >= ELMO_DEVICES) {
	       if (debug) printk(KERN_DEBUG "Rejecting command for device %d, bad target\n", device);
	       tmp->result = DID_BAD_TARGET << 16;
	       tmp->scsi_done(tmp);
	       done = 0;
	       tmp = prev;                        
	       continue;
	     }
	     /*
	       Initialize the buffer pointers
	     */
	     initialize_SCp(tmp);
	     /*
	       Connect this request to the handler and mark the handler busy with
	       a request not yet successfully initiated.
	     */
	     hostdata->connected[deviceDispatcher[device]] = tmp;
	     hostdata->busy[deviceDispatcher[device]] = 1;
	     hostdata->pending[deviceDispatcher[device]] = 1;
                    
	     /* When removing an item, we want to hook up the next item to our
	        prev item (not ourself).  At loop iteration time, this will
	        happen and prevents the removed item from being
	        unintentionally left in the list.
	     */
	     tmp = prev;
	   }
	 }
       /*
	 OK, we've walked the queue, and connected up any requests for
	 dispatchers which were previously idle.  Now, kick off any
	 dispatchers which have uninitiated requests.
       */
       write_unlock_irqrestore(&hostdata->queueLock, iflags);
       for (handler = 0; handler < ELMO_HANDLERS; handler++) {
	 /*
	   Call the handler for this [pseudo]target
	 */
	 if (hostdata->pending[handler]) {
	   int result;
	   hostdata->pending[handler] = 0;
	   tmp = (Scsi_Cmnd *) hostdata->connected[handler];
	   result = (*dispatchers[handler])(instance, tmp, tmp->target);
	   /*
	     Dispatchers return 0 if request accepted and in progress, 1 if 
	     request was completed immediately and should be marked as done,
	     and -1 if the request couldn't be accepted and should be requeued
	     on the head of the list.
	   */
	   if (result != 0) {
	     hostdata->connected[handler] = NULL;
	     hostdata->busy[handler] = 0;
	     if (result < 0) {
	       write_lock_irqsave(&hostdata->queueLock, iflags);
	       LIST(tmp, hostdata->issue_queue);
	       tmp->host_scribble = (unsigned char *) hostdata->issue_queue;
	       hostdata->issue_queue = tmp;
	       write_unlock_irqrestore(&hostdata->queueLock, iflags);
	     } else {
	       if (debug) printk(KERN_INFO "elmo: elmo_main calls done on %p\n", tmp);
	       tmp->scsi_done(tmp);
	       done = 0;
	     }
	   }
	 }
       }
       done &= elmo_run_workers(instance);
     }
   } while (!done);
   /* Exit lock held */
   clear_bit(0, &main_running);

}

static void show_elmo_intr(unsigned long combinedISR)
{
  const int numIntrs = 32;
  const char *intrs[] = {
    " EBI_TO_ESS_NE",
    " EBI_TO_ESS_WM",
    " EBI_TO_ESS_FF",
    " BROADCOM_DMA",
    " ESS_TO_EBI_NE",
    " ESS_TO_EBI_WM",
    " TIMER",
    " RFU_BIT_7",
    " EBI_TO_IDE_NE",
    " EBI_TO_IDE_WM",
    " EBI_TO_IDE_FF",
    " RFU_BIT_11",
    " IDE_TO_EBI_NE",
    " IDE_TO_EBI_WM",
    " RFU_BIT_14",
    " RFU_BIT_14",
    " BUS_TIMEOUT",
    " ELMO_ASIC_DMA_RD_DONE",
    " ELMO_ASIC_DMA_WR_DONE",
    " RFU_BIT_3",
    " RFU_BIT_4",
    " DSK_RST_REQ",
    " DSK_CMD_REQ",
    " INT_FROM_ESS",
    " EBI_TO_ESS_OVR",
    " EBI_TO_ESS_UDR",
    " ESS_TO_EBI_OVR",
    " ESS_TO_EBI_UDR",
    " EBI_TO_IDE_OVR",
    " EBI_TO_IDE_UDR",
    " IDE_TO_EBI_OVR",
    " IDE_TO_EBI_UDR" };
  int i;
  if (!combinedISR) {
    return;
  }
  printk(KERN_INFO "elmo: interrupt");
  for (i = numIntrs - 1; i >= 0; i--) {
    if (combinedISR & (1 << i)) {
      printk(intrs[i]);
    }
  }
  printk("\n");
}

/**
 *
 *      Broadcom "end of DMA" interrupt handlers
 *
 */

static void broadcom_intr(int irq, int *useFlag, unsigned long softInterrupt)
{
  unsigned long flags;
  struct Scsi_Host *instance;
  unsigned long interruptsTaken;

  save_and_cli(flags);

  if (debug) printk(KERN_INFO "elmo: Broadcom DMA interrupt %d triggered!\n", irq);

  disable_irq(irq);

  interruptsTaken = 0;
  for (instance = first_instance;
       instance && instance->hostt == the_template;
       instance = instance->next) {
    interruptsTaken |= elmo_interrupt_for_workers(instance, softInterrupt);
  }

  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);

  if (interruptsTaken) {
      *useFlag = 0;
  }

  restore_flags(flags);
  
  if (!interruptsTaken) {
    printk(KERN_ERR "Oops!  No worker took Broadcom DMA interrupt %d!\n", irq);
  }
}

static void broadcom_isb_intr(int irq, void *dev_id, struct pt_regs *regs)
{
  (void) dev_id;
  (void) regs;
  *BCM_DMA_CH1_INT_STAT = 0x4;
  *BCM_DMA_CH2_INT_STAT = 0x4;
  barrier();
  broadcom_intr(irq, &elmo_broadcom_isb_enabled, ELMO_INT_ISB_DMA);
}

/**
 * 	elmo_intr	-	Elmo interrupt handler
 *
 *	Handle interrupts.
 *
 *	Locks: caller must hold the io_request lock.
 */

static void elmo_intr(int irq, void *dev_id, struct pt_regs *regs) {
  struct Scsi_Host *instance;

  unsigned long interruptsTaken;
  unsigned long wakeups;
  unsigned short bridgeISR, bridgeFISR;

  /*
    Do the magic here.  Read out and clear the interrupts in the main ISR
    which triggered us.  Also, read out the FIFO status register
    conditions which triggered us, and mask them off so that they
    do not trigger again unless some worker unmasks them.
  */

  bridgeISR = ELMO_BRIDGE->isr & ELMO_BRIDGE->imr;
  barrier();
  ELMO_BRIDGE->isr = bridgeISR; /* clear everything we can... */
  barrier();

  bridgeFISR = ELMO_BRIDGE->fsr & ELMO_BRIDGE->fmr;
  barrier();
  ELMO_BRIDGE->fmr &= ~bridgeFISR;  /* mask off the things which triggered */
  barrier();

  /*
    Combine the two 16-bit interrupt sets in a single 32-bit set which
    defines the 'virtual' interrupts we need to process.
  */

  wakeups = (((unsigned long) bridgeISR) << 16) | bridgeFISR;

  PLOG( PLOG_ELMO_RPC_ELMO_INTERRUPT, sizeof(wakeups), &wakeups );

  if (debug) {
    show_elmo_intr(wakeups);
  }

  /*
    Kick awake any workers who were waiting for any of the interrupts
    that we just noticed, and then tickle the bottom-half task
  */

  if (wakeups) {

    interruptsTaken = 0;

    for (instance = first_instance;
	 instance && instance->hostt == the_template;
	 instance = instance->next) {
      interruptsTaken |= elmo_interrupt_for_workers(instance, wakeups);
    }

    if (interruptsTaken & ELMO_INT_DMA_RD_IRQ) {
        elmo_asic_read_dma_enabled = 0;
        *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0x1000U);
    }else if (interruptsTaken & ELMO_INT_DMA_WR_IRQ) {
        elmo_asic_write_dma_enabled = 0;
        *(volatile unsigned short *)KSEG1ADDR(ELMO_DMA_EN)&=~(0x2000U);
    }
    barrier();

    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);

    if (interruptsTaken != wakeups) {
      printk(KERN_ERR "elmo: OOPS:  Unclaimed virtual interrupt 0x%08lX\n",
	     wakeups & ~interruptsTaken);
      show_elmo_intr(wakeups & ~interruptsTaken);
    }
  }
	   
}

/**
 *	do_elmo_intr
 *	@irq: interrupt number
 *	@dev_id: device info
 *	@regs: registers (unused)
 *
 *	Takes the io_request_lock and invokes the Elmo interrupt
 *	handler code
 *
 *	Locks: takes and releases the io_request lock
 */

static void do_elmo_intr(int irq, void *dev_id, struct pt_regs *regs) {
  unsigned long flags;

  volatile unsigned short imr = 0;
  volatile unsigned short fmr = 0;

  spin_lock_irqsave(&io_request_lock, flags);
  elmo_intr(irq, dev_id, regs);
  spin_unlock_irqrestore(&io_request_lock, flags);

  /* clear the interrupt mask register and set it to the same value. We need to
   * do this because the 7317 does not support level triggered interrupts on
   * GPIOs. We have to clear and set the ASIC interrupt mask register to force
   * an edge transition on the GPIO line, if an interrupt is pending. This will
   * hold the interrupt line active, so the 7317 will detect the interrupt and
   * service it. Otherwise, there is a possibility for a pending interrupt to
   * be missed
   */
  imr = ELMO_BRIDGE->imr;
  barrier();
  ELMO_BRIDGE->imr = 0;
  barrier();
  ELMO_BRIDGE->imr = imr;

  /* clear the fifo interrupt mask register and set it to the same value. 
   * We need to do this because the 7317 does not support level triggered 
   * interrupts on GPIOs. We have to clear and set the ASIC interrupt mask 
   * register to force an edge transition on the GPIO line, if an interrupt
   * is pending. This will hold the interrupt line active, so the 7317 will 
   * detect the interrupt and service it. Otherwise, there is a possibility 
   * for a pending interrupt to be missed
   */
  fmr = ELMO_BRIDGE->fmr;
  barrier();
  ELMO_BRIDGE->fmr = 0;
  barrier();
  ELMO_BRIDGE->fmr = fmr;
  barrier();
}

/**
   elmo_abort - abort a command.
 */

static int elmo_abort(Scsi_Cmnd * cmd)
{
   struct Scsi_Host *instance;
   struct elmo_hostdata *hostdata;
   struct list_head *this;
   struct elmo_worker *worker;
   Scsi_Cmnd *tmp, *prev, *next;
   int handler;
   unsigned long iflags;
   /*
     Scan the driver instances (should only be one, really).
     For each instance, check for four cases:
     [1] Command is on the issue queue.  Easy case... dequeue it and
         it's aborted.
     [2] Command is connected to a device, and pending (i.e. hasn't
         actually been dispatched yet).  Just as easy... disconnect it,
         un-pend the device, and it's aborted.
     [3] Command is connected to the device and not pending.  Can't
         abort it in this case - return a SCSI_ABORT_NOT_RUNNING.
     [4] Command isn't in the issue queue, isn't connected, and there's
         a worker which owns it.  We tell the worker to abort the
         command, kick the worker awake, and return SCSI_ABORT_PENDING.
   */

   printk(KERN_ERR "elmo: in elmo_abort()\n");

   printk(KERN_ERR "elmo:  ctl %04X\n", ELMO_BRIDGE->control);
   printk(KERN_ERR "elmo:  isr %04X\n", ELMO_BRIDGE->isr);
   printk(KERN_ERR "elmo:  imr %04X\n", ELMO_BRIDGE->imr);
   printk(KERN_ERR "elmo:  fsr %04X\n", ELMO_BRIDGE->fsr);
   printk(KERN_ERR "elmo:  fmr %04X\n", ELMO_BRIDGE->fmr);

   elmo_autodebug_on();

   for (instance = first_instance;
	instance && instance->hostt == the_template;
	instance = instance->next) {
     hostdata = (struct elmo_hostdata *) instance->hostdata;
     write_lock_irqsave(&hostdata->queueLock, iflags);
     for (tmp = (Scsi_Cmnd *) hostdata->issue_queue, prev = NULL;
	  tmp;
	  prev = tmp, tmp = next) {
       next = (Scsi_Cmnd *) tmp->host_scribble;
       if (tmp == cmd) {
	 /* found it in the issue queue.  Dequeue and abort. */
	 if (prev) {
	   REMOVE(prev, prev->host_scribble, tmp, tmp->host_scribble);
	   prev->host_scribble = tmp->host_scribble;
	 } else {
	   REMOVE(-1, hostdata->issue_queue, tmp, tmp->host_scribble);
	   hostdata->issue_queue = (Scsi_Cmnd *) tmp->host_scribble;
	 }
	 tmp->host_scribble = NULL;
	 tmp->result = DID_ABORT << 16;
	 write_unlock_irqrestore(&hostdata->queueLock, iflags);
	 printk(KERN_INFO "elmo: aborted a command in the issue queue\n");
	 return SCSI_ABORT_SUCCESS;
       }
     }
     for (handler = 0; handler < ELMO_HANDLERS; handler++) {
       tmp = (Scsi_Cmnd *) hostdata->connected[handler];
       if (tmp == cmd) {
	 if (hostdata->pending[handler]) {
	   hostdata->connected[handler] = NULL;
	   hostdata->pending[handler] = 0;
	   tmp->result = DID_ABORT << 16;
	   write_unlock_irqrestore(&hostdata->queueLock, iflags);
	   printk(KERN_INFO "elmo: aborted a command which was pending\n");
	   return SCSI_ABORT_SUCCESS;
	 } else {
	   write_unlock_irqrestore(&hostdata->queueLock, iflags);
	   printk(KERN_INFO "elmo: can't abort a connected command\n");
	   return SCSI_ABORT_NOT_RUNNING;
	 }
       }
     }
     list_for_each(this,&hostdata->workers) {
       worker = (struct elmo_worker *) this;
       if (worker->scsiRequest == cmd) {
	 worker->aborting = 1;
	 worker->go = 1;
	 elmo_worker_disarm_timer(worker);
	 queue_task (&elmo_task_struct, &tq_immediate);
	 mark_bh (IMMEDIATE_BH);
	 write_unlock_irqrestore(&hostdata->queueLock, iflags);
	 printk(KERN_INFO "elmo: aborting a command in progress\n");
	 return SCSI_ABORT_PENDING;
       }
     }
     write_unlock_irqrestore(&hostdata->queueLock, iflags);
   }
   printk(KERN_INFO "elmo: didn't find that command, can't abort it\n");
   return SCSI_ABORT_NOT_RUNNING;
}

/**
   elmo_reset - reset device or adapter.  The command passed in is a
   dummy (usually) used only to identify the device we wish
   to reset.

   The reset functionality here is as quirky as most of the
   other stuff in this driver. 
 */

static int elmo_reset(Scsi_Cmnd * cmd, unsigned int reset_flags)
{
   struct Scsi_Host *instance;
   struct elmo_hostdata *hostdata;
   unsigned long iflags;

   (void) reset_flags;
   printk(KERN_ERR "elmo: in elmo_reset()\n");

   instance = cmd->host;
   hostdata = (struct elmo_hostdata *) instance->hostdata;

   save_and_cli(iflags);
   hostdata->wasReset = 1;
   queue_task (&elmo_task_struct, &tq_immediate);
   mark_bh (IMMEDIATE_BH);
   restore_flags(iflags);
   return SCSI_RESET_SUCCESS;
}

/**
   elmo_info - get info on a controller.
 */

static const char *elmo_info(struct Scsi_Host *instance)
{
  (void) instance;
  if (debug) printk(KERN_DEBUG "In elmo_info()\n");
  return "ELMO controller";
}

/**
   elmo_proc_info - get info on a controller.
 */

#define PRINTP(x) len += sprintf(buffer+len, x)
#define ANDP ,

static int elmo_proc_info(char *buffer, char **start, off_t offset,
			int length, int hostno, int inout)
{
  int len = 0;
  struct Scsi_Host *scsi_ptr;
  struct elmo_hostdata *hostdata;
  unsigned long flags;
  (void) start;
  (void) offset;
  (void) length;
  (void) hostno;
  (void) inout;
  if (debug) printk(KERN_DEBUG "In elmo_proc_info()\n");
  save_and_cli(flags);
  for (scsi_ptr = first_instance; scsi_ptr; scsi_ptr = scsi_ptr->next) {
    if (scsi_ptr->host_no == hostno) {
      break;
    }
  }
  hostdata = (struct elmo_hostdata *) scsi_ptr->hostdata;
  PRINTP("SCSI host number %d : %s\n" ANDP scsi_ptr->host_no ANDP scsi_ptr->hostt->name);
  if (hostdata->pldLoaded) {
    PRINTP("Bridge : UP, version %02X\n" ANDP (0xFF & ELMO_BRIDGE->revision));
  } else {
    PRINTP("Bridge : DOWN\n");
  }
  *start = buffer + offset;
  len -= offset;
  if (len > length)
    len = length;
  restore_flags(flags);
  return len;
}


/**
 *	elmo_queue_command 		-	queue a command
 *	@cmd: SCSI command
 *	@done: completion handler
 *
 *      cmd is added to the per instance issue_queue, with minor 
 *      twiddling done to the host specific fields of cmd.  If the 
 *      main coroutine is not running, it is restarted.
 *
 *	Locks: io_request lock held by caller. Called functions drop and
 *	retake this lock. Called functions take dma lock.
 */

int elmo_queue_command(Scsi_Cmnd * cmd, void (*done) (Scsi_Cmnd *)) {
	struct Scsi_Host *instance = cmd->host;
	struct elmo_hostdata *hostdata = (struct elmo_hostdata *) instance->hostdata;
	Scsi_Cmnd *tmp;
	unsigned long iflags;


	/* 
	 * We use the host_scribble field as a pointer to the next command  
	 * in a queue 
	 */

	cmd->host_scribble = NULL;
	cmd->scsi_done = done;
	cmd->result = 0;

	if (debug) printk(KERN_DEBUG "elmo: queue_command %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
	 cmd->cmnd[0], cmd->cmnd[1], cmd->cmnd[2], cmd->cmnd[3], 
	 cmd->cmnd[4], cmd->cmnd[5], cmd->cmnd[6], cmd->cmnd[7], 
	 cmd->cmnd[8], cmd->cmnd[9], cmd->cmnd[10], cmd->cmnd[11]);

	/* 
	 * Insert the cmd into the issue queue. Note that REQUEST SENSE 
	 * commands are added to the head of the queue since any command will
	 * clear the contingent allegiance condition that exists and the 
	 * sense data is only guaranteed to be valid while the condition exists.
	 */

	write_lock_irqsave(&hostdata->queueLock, iflags);

	if (!(hostdata->issue_queue) || (cmd->cmnd[0] == REQUEST_SENSE)) {
		LIST(cmd, hostdata->issue_queue);
		cmd->host_scribble = (unsigned char *) hostdata->issue_queue;
		hostdata->issue_queue = cmd;
	} else {
		for (tmp = (Scsi_Cmnd *) hostdata->issue_queue; tmp->host_scribble; tmp = (Scsi_Cmnd *) tmp->host_scribble);
		LIST(cmd, tmp);
		tmp->host_scribble = (unsigned char *) cmd;
	}

	write_unlock_irqrestore(&hostdata->queueLock, iflags);

	/*
	  Kick the runner task awake if it's not already so...
	*/
	queue_task (&elmo_task_struct, &tq_immediate);
	mark_bh (IMMEDIATE_BH);

	return 0;
}

/*
 * Ioctl handler
 */

static int elmo_ioctl(Scsi_Device *dev, int cmd, void *arg)
{
   struct Scsi_Host *host = dev->host;
   unsigned int id = dev->id;
   if (debug) printk(KERN_DEBUG "In elmo_ioctl()\n");
   if (ioctlHandlers[id]) {
     return (*ioctlHandlers[id])(host, dev, cmd, arg);
   } 
   return -EINVAL;
}

/*
 * Handlers for read/write/ioctl on the Elmo low-latency interface.
 */

ssize_t elmo_ll_read (struct file *filp, char *b, size_t t, loff_t *l)
{
  struct elmo_worker *worker;
  unsigned int flags;
  int ret;
  if (t == 0) {
    return 0;
  } else if (t > ELMO_READBUFFER_SIZE) {
    t = ELMO_READBUFFER_SIZE;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_ATOMIC);
  if (debug > 1) printk(KERN_INFO "elmo: create worker for reading\n");
  if (!worker) {
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++thehostdata->workerSN;
  if (debug > 1) printk(KERN_INFO "elmo: allocating %d-byte buffer\n", t);
  worker->dataBuffer = kmalloc(t, GFP_ATOMIC);
  if (!worker->dataBuffer) {
    kfree(worker);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  if (debug > 1) printk(KERN_INFO "elmo: queueing worker %d for read\n",
			worker->sn);
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_READER);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker, &thehostdata->workers);
  restore_flags(flags);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  if (debug > 1) printk(KERN_INFO "elmo: waiting on worker\n");
  ret = wait_event_interruptible(readerWaitQueue, worker->readCompleted);
  if (ret == 0) {
    copy_to_user (b, worker->dataBuffer, worker->dataBufferCount);
    ret = worker->dataBufferCount;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  return ret;
}

ssize_t elmo_ll_readv (struct file *filp, const struct iovec *iov, 
		       unsigned long count, loff_t *l)
{
  struct elmo_worker *worker;
  unsigned int flags;
  size_t t;
  size_t have, offset, take;
  int i;
  int ret = 0;
  t = 0;
  if (debug > 1) printk(KERN_INFO "elmo: summing iovec for readv\n");
  for (i = 0; i < count; i++) {
    if (verify_area(VERIFY_WRITE, iov[i].iov_base, iov[i].iov_len)) {
      return -EFAULT;
    }
    t += iov[i].iov_len;
  }
  if (debug > 1) printk(KERN_INFO "elmo: readv total size is %d\n", t);
  if (t == 0) {
    return 0;
  } else if (t > ELMO_READBUFFER_SIZE) {
    t = ELMO_READBUFFER_SIZE;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_ATOMIC);
  if (!worker) {
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++thehostdata->workerSN;
  worker->dataBuffer = kmalloc(t, GFP_ATOMIC);
  if (!worker->dataBuffer) {
    kfree(worker);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_READER);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker, &thehostdata->workers);
  restore_flags(flags);

  PLOG( PLOG_ELMO_RPC_READV_STARTED, 0, 0 );

  run_main();
  have = worker->dataBufferCount;

  PLOG( PLOG_ELMO_RPC_READV_HAS, sizeof( have ), &have );

  if ( have == 0 ) {
      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);
      ret = wait_event_interruptible(readerWaitQueue, worker->readCompleted);
  }
  if (debug > 1) printk(KERN_INFO "readv back from wait, code %d\n", ret);
  if (ret != 0) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);
    return ret;
  }
  offset = 0;
  have = worker->dataBufferCount;

  PLOG( PLOG_ELMO_RPC_READV_HAS2, sizeof( have ), &have );

  if (debug > 1) printk(KERN_INFO "readv has %d available\n", have);
  for (i = 0; i < count && have > 0; i++) {
    take = have;
    if (take > iov[i].iov_len) {
      take = iov[i].iov_len;
    }
    copy_to_user(iov[i].iov_base, &worker->dataBuffer[offset], take);
    offset += take;
    have -= take;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  return offset;
}

ssize_t elmo_ll_write_common (struct file *filp, const char *b, size_t t,
			      loff_t *l, int defer)
{
  struct elmo_worker *worker;
  unsigned int flags;
  int ret;
  if (t == 0) {
    return 0;
  } else if (t > 4095) {
    return -EINVAL;
  }
  if (debug > 1) printk(KERN_INFO "elmo: writer sem now %d\n",
		    sem_getcount(&thehostdata->writeSem));
  ret = down_interruptible(&thehostdata->writeSem);
  if (ret) {
    return ret;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_ATOMIC);
  if (!worker) {
    up(&thehostdata->writeSem);
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++thehostdata->workerSN;
  worker->holdingSemaphore = 1;
  worker->dataBuffer = kmalloc(t, GFP_ATOMIC);
  if (debug > 1) printk(KERN_INFO "elmo: worker %d downs sem to %d\n", worker->sn,
		    sem_getcount(&thehostdata->writeSem));
  if (!worker->dataBuffer) {
    kfree(worker);
    up(&thehostdata->writeSem);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  copy_from_user (worker->dataBuffer, b, t);
  save_and_cli(flags);
  if (defer) {
    elmo_worker_next_state(worker, ELMO_WORKER_GATE_WRITER);
    thehostdata->resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
  } else {
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  }
  list_add_tail((struct list_head *) worker, &thehostdata->workers);
  if (!defer) {
    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);
  }
  restore_flags(flags);
  return t;
}

ssize_t elmo_ll_write (struct file *filp, const char *b, size_t t, loff_t *l)
{
  return elmo_ll_write_common (filp, b, t, l, 0);
}

ssize_t elmo_ll_write_deferred (struct file *filp, const char *b, size_t t,
				loff_t *l)
{
  return elmo_ll_write_common (filp, b, t, l, 1);
}

ssize_t elmo_ll_writev_common (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *l, int defer)
{
  struct elmo_worker *worker;
  size_t t;
  size_t offset, take;
  int i;
  int take_count = count;
  unsigned int flags;
  int ret;
  t = 0;
  for (i = 0; i < count; i++) {
    if (t + iov[i].iov_len > 4095) {
        take_count = i;
        break;
    }
    t += iov[i].iov_len;
  }
  if (t == 0) {
    return 0;
  } else if (t > 4095) {
    return -EINVAL;
  }
  if (debug > 1) printk(KERN_INFO "elmo: writer sem now %d\n",
		    sem_getcount(&thehostdata->writeSem));
  ret = down_interruptible(&thehostdata->writeSem);
  if (ret) {
    return ret;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_ATOMIC);
  if (!worker) {
    up(&thehostdata->writeSem);
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++thehostdata->workerSN;
  worker->holdingSemaphore = 1;
  if (debug > 1) printk(KERN_INFO "elmo: worker %d downs sem to %d\n", worker->sn,
		    sem_getcount(&thehostdata->writeSem));
  worker->dataBuffer = kmalloc(t, GFP_ATOMIC);
  if (!worker->dataBuffer) {
    kfree(worker);
    up(&thehostdata->writeSem);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  offset = 0;
  for (i = 0; i < take_count; i++) {
    take = iov[i].iov_len;
    copy_from_user(&worker->dataBuffer[offset], iov[i].iov_base, take);
    offset += take;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  save_and_cli(flags);
  if (defer) {
    elmo_worker_next_state(worker, ELMO_WORKER_GATE_WRITER);
    thehostdata->resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
  } else {
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  }
  list_add_tail((struct list_head *) worker, &thehostdata->workers);
  restore_flags(flags);

  run_main();

  save_and_cli(flags);
  if (!defer) {
      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);
  }
  restore_flags(flags);

  PLOG( PLOG_ELMO_RPC_LL_WRITE_DONE, 0, 0);

  return t;
}

ssize_t elmo_ll_writev (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *l) 
{
  return elmo_ll_writev_common (filp, iov, count, l, 0);
}
  

ssize_t elmo_ll_writev_deferred (struct file *filp, const struct iovec *iov, 
				 unsigned long count, loff_t *l) 
{
  return elmo_ll_writev_common (filp, iov, count, l, 1);
}

int elmo_ll_ioctl (struct inode *i, struct file *f, unsigned int cd,
		   unsigned long param)
{
  printk(KERN_INFO "elmo: got ioctl code 0x%08X param 0x%08lX\n", cd, param);
  return -EINVAL;
}

int elmo_ll_open (struct inode *i, struct file *filp)
{
  int type = MINOR(i->i_rdev);

  if (type > 0) {
    filp->f_op = &elmo_ll_deferred_fops;
  }

  return 0;

}

#ifdef MODULE

static Scsi_Host_Template driver_template = ELMO;

#include <linux/module.h>
#include "scsi_module.c"

#endif

EXPORT_SYMBOL(ELMOReadBridge);
EXPORT_SYMBOL(ELMOWriteBridge);
EXPORT_SYMBOL(ELMOWriteBridgeMasked);
